#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Ref:
# Chem Phys Lett, 256, 454
# J. Mol. Struct. THEOCHEM, 914, 3
#

from functools import reduce
import numpy
from pyscf import lib
from pyscf.lib import linalg_helper
from pyscf.lib import logger
from pyscf.tdscf import rhf
from pyscf.tdscf._lr_eig import eigh as lr_eigh, eig as lr_eig
from pyscf.pbc import scf
from pyscf.pbc.tdscf.rhf import TDBase
from pyscf.pbc.scf import _response_functions  # noqa
from pyscf.pbc.lib.kpts_helper import is_gamma_point, get_kconserv_ria, conj_mapping
from pyscf.pbc.df.df_ao2mo import warn_pbc2d_eri
from pyscf.pbc import df as pbcdf
from pyscf.data import nist
from pyscf import __config__

REAL_EIG_THRESHOLD = getattr(__config__, 'pbc_tdscf_rhf_TDDFT_pick_eig_threshold', 1e-3)

def get_ab(mf, kshift=0):
    r'''A and B matrices for TDDFT response function.

    A[i,a,j,b] = \delta_{ab}\delta_{ij}(E_a - E_i) + (ai||jb)
    B[i,a,j,b] = (ai||bj)

    Ref: Chem Phys Lett, 256, 454

    Kwargs:
        kshift : integer
            The index of the k-point that represents the transition between
            k-points in the excitation coefficients.
    '''
    cell = mf.cell
    mo_energy = scf.addons.mo_energy_with_exxdiv_none(mf)
    mo = numpy.asarray(mf.mo_coeff)
    mo_occ = numpy.asarray(mf.mo_occ)
    kpts = mf.kpts
    nkpts, nao, nmo = mo.shape
    noccs = numpy.count_nonzero(mo_occ==2, axis=1)
    nocc = noccs[0]
    nvir = nmo - nocc
    assert all(noccs == nocc)
    orbo = mo[:,:,:nocc]
    orbv = mo[:,:,nocc:]

    kconserv = get_kconserv_ria(cell, kpts)[kshift]
    e_ia = numpy.asarray(_get_e_ia(mo_energy, mo_occ, kconserv)).astype(mo.dtype)
    a = numpy.diag(e_ia.ravel()).reshape(nkpts,nocc,nvir,nkpts,nocc,nvir)
    b = numpy.zeros_like(a)
    weight = 1./nkpts

    def add_hf_(a, b, hyb=1):
        eri = mf.with_df.ao2mo_7d([mo,orbo,mo,mo], kpts)
        eri *= weight
        eri = eri.reshape(nkpts,nkpts,nkpts,nmo,nocc,nmo,nmo)
        for ki, ka in enumerate(kconserv):
            for kj, kb in enumerate(kconserv):
                a[ki,:,:,kj] += numpy.einsum('aijb->iajb', eri[ka,ki,kj,nocc:,:,:nocc,nocc:]) * 2
                a[ki,:,:,kj] -= numpy.einsum('jiab->iajb', eri[kj,ki,ka,:nocc,:,nocc:,nocc:]) * hyb

            for kb, kj in enumerate(kconserv):
                b[ki,:,:,kj] += numpy.einsum('aibj->iajb', eri[ka,ki,kb,nocc:,:,nocc:,:nocc]) * 2
                b[ki,:,:,kj] -= numpy.einsum('ajbi->iajb', eri[ka,kj,kb,nocc:,:,nocc:,:nocc]) * hyb

    if isinstance(mf, scf.hf.KohnShamDFT):
        ni = mf._numint
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, cell.spin)

        add_hf_(a, b, hyb)
        if omega != 0:  # For RSH
            raise NotImplementedError

        xctype = ni._xc_type(mf.xc)
        dm0 = mf.make_rdm1(mo, mo_occ)
        make_rho = ni._gen_rho_evaluator(cell, dm0, hermi=1, with_lapl=False)[0]
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)
        cmap = conj_mapping(cell, kpts)

        if xctype == 'LDA':
            ao_deriv = 0
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho = make_rho(0, ao, mask, xctype)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc[0,0] * weight

                rho_o = lib.einsum('krp,kpi->kri', ao, orbo)
                rho_v = lib.einsum('krp,kpi->kri', ao, orbv)
                rho_ov = numpy.einsum('kri,kra->kria', rho_o, rho_v)
                rho_vo = rho_ov.conj()[cmap]
                w_vo = numpy.einsum('kria,r->kria', rho_vo, wfxc) * (2/nkpts)
                a += lib.einsum('kria,lrjb->kialjb', w_vo, rho_ov)
                b += lib.einsum('kria,lrjb->kialjb', w_vo, rho_vo)

        elif xctype == 'GGA':
            ao_deriv = 1
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho = make_rho(0, ao, mask, xctype)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc * weight
                rho_o = lib.einsum('kxrp,kpi->kxri', ao, orbo)
                rho_v = lib.einsum('kxrp,kpi->kxri', ao, orbv)
                rho_ov = numpy.einsum('kxri,kra->kxria', rho_o, rho_v[:,0])
                rho_ov[:,1:4] += numpy.einsum('kri,kxra->kxria', rho_o[:,0], rho_v[:,1:4])
                rho_vo = rho_ov.conj()[cmap]
                w_vo = numpy.einsum('xyr,kxria->kyria', wfxc, rho_vo) * (2/nkpts)
                a += lib.einsum('kxria,lxrjb->kialjb', w_vo, rho_ov)
                b += lib.einsum('kxria,lxrjb->kialjb', w_vo, rho_vo)

        elif xctype == 'HF':
            pass

        elif xctype == 'NLC':
            raise NotImplementedError('NLC')

        elif xctype == 'MGGA':
            ao_deriv = 1
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho = make_rho(0, ao, mask, xctype)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc * weight
                rho_o = lib.einsum('kxrp,kpi->kxri', ao, orbo)
                rho_v = lib.einsum('kxrp,kpi->kxri', ao, orbv)
                rho_ov = numpy.einsum('kxri,kra->kxria', rho_o, rho_v[:,0])
                rho_ov[:,1:4] += numpy.einsum('kri,kxra->kxria', rho_o[:,0], rho_v[:,1:4])
                tau_ov = numpy.einsum('kxri,kxra->kria', rho_o[:,1:4], rho_v[:,1:4]) * .5
                rho_ov = numpy.concatenate([rho_ov, tau_ov[:,numpy.newaxis]], axis=1)
                rho_vo = rho_ov.conj()[cmap]
                w_vo = numpy.einsum('xyr,kxria->kyria', wfxc, rho_vo) * (2/nkpts)
                a += lib.einsum('kxria,lxrjb->kialjb', w_vo, rho_ov)
                b += lib.einsum('kxria,lxrjb->kialjb', w_vo, rho_vo)
    else:
        add_hf_(a, b)

    return a, b

class KTDBase(TDBase):
    '''
    Attributes:
        kshift_lst : list of integers
            Each element in the list is the index of the k-point that
            represents the transition between k-points in the excitation
            coefficients. For excitation amplitude T_{ai}[k] at point k,
            the kshift connects orbital i at k and orbital a at k+kshift
    '''

    conv_tol = getattr(__config__, 'pbc_tdscf_rhf_TDA_conv_tol', 1e-4)

    _keys = {'kshift_lst'}

    def __init__(self, mf, kshift_lst=None):
        assert isinstance(mf, scf.khf.KSCF)
        TDBase.__init__(self, mf)
        warn_pbc2d_eri(mf)

        if kshift_lst is None: kshift_lst = [0]
        self.kshift_lst = kshift_lst

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('\n')
        log.info('******** %s for %s ********',
                 self.__class__, self._scf.__class__)
        if self.singlet is None:
            log.info('nstates = %d', self.nstates)
        elif self.singlet:
            log.info('nstates = %d singlet', self.nstates)
        else:
            log.info('nstates = %d triplet', self.nstates)
        log.info('deg_eia_thresh = %.3e', self.deg_eia_thresh)
        log.info('kshift_lst = %s', self.kshift_lst)
        log.info('wfnsym = %s', self.wfnsym)
        log.info('conv_tol = %g', self.conv_tol)
        log.info('eigh lindep = %g', self.lindep)
        log.info('eigh level_shift = %g', self.level_shift)
        log.info('eigh max_cycle = %d', self.max_cycle)
        log.info('chkfile = %s', self.chkfile)
        log.info('max_memory %d MB (current use %d MB)',
                 self.max_memory, lib.current_memory()[0])
        if not self._scf.converged:
            log.warn('Ground state SCF is not converged')
        log.info('\n')

    def check_sanity(self):
        TDBase.check_sanity(self)
        mf = self._scf
        if any(k != 0 for k in self.kshift_lst):
            if mf.rsjk is not None or not isinstance(mf.with_df, pbcdf.df.DF):
                logger.error(self, 'Solutions with non-zero kshift for %s are '
                             'only supported by GDF/RSDF')
                raise NotImplementedError

    def _finalize(self):
        '''Hook for dumping results and clearing up the object.'''
        for k,kshift in enumerate(self.kshift_lst):
            if not all(self.converged[k]):
                logger.note(self, 'kshift = %d  TD-SCF states %s not converged.',
                            kshift, [i for i, x in enumerate(self.converged[k]) if not x])
            logger.note(self, 'kshift = %d  Excited State energies (eV)\n%s',
                        kshift, self.e[k] * nist.HARTREE2EV)
        return self

    get_nto = lib.invalid_method('get_nto')

class TDA(KTDBase):

    @lib.with_doc(get_ab.__doc__)
    def get_ab(self, mf=None, kshift=0):
        if mf is None: mf = self._scf
        return get_ab(mf, kshift)

    def gen_vind(self, mf=None, kshift=0):
        '''Compute Ax

        Kwargs:
            kshift : integer
                The index of the k-point that represents the transition between
                k-points in the excitation coefficients.
        '''
        assert mf is self._scf
        singlet = self.singlet
        kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        nkpts = len(mo_occ)
        nao, nmo = mo_coeff[0].shape
        occidx = [mo_occ[k]==2 for k in range(nkpts)]
        viridx = [mo_occ[k]==0 for k in range(nkpts)]
        orbo = [mo_coeff[k][:,occidx[k]] for k in range(nkpts)]
        orbv = [mo_coeff[k][:,viridx[k]] for k in range(nkpts)]
        dtype = numpy.result_type(*mo_coeff)
        e_ia = _get_e_ia(scf.addons.mo_energy_with_exxdiv_none(mf), mo_occ, kconserv)
        hdiag = numpy.hstack([x.ravel() for x in e_ia])

        mem_now = lib.current_memory()[0]
        max_memory = max(2000, self.max_memory*.8-mem_now)
        vresp = self.gen_response(singlet=singlet, hermi=0, max_memory=max_memory)

        def vind(zs):
            nz = len(zs)
            z1s = [_unpack(z, mo_occ, kconserv) for z in zs]
            dms = numpy.empty((nz,nkpts,nao,nao), dtype=dtype)
            for i in range(nz):
                for k, kp in enumerate(kconserv):
                    # *2 for double occupancy
                    dm1 = z1s[i][k] * 2
                    dms[i,kp] = lib.einsum('ov,pv,qo->pq', dm1, orbv[kp], orbo[k].conj())

            with lib.temporary_env(mf, exxdiv=None):
                v1ao = vresp(dms, kshift)
            v1s = []
            for i in range(nz):
                dm1 = z1s[i]
                v1 = [None] * nkpts
                for k, kp in enumerate(kconserv):
                    v1mo = lib.einsum('pq,qo,pv->ov', v1ao[i,kp], orbo[k], orbv[kp].conj())
                    v1mo += e_ia[k] * dm1[k]
                    v1[k] = v1mo.ravel()
                v1s.append( numpy.concatenate(v1) )
            return numpy.stack(v1s)
        return vind, hdiag

    def get_init_guess(self, mf, kshift, nstates=None):
        if nstates is None: nstates = self.nstates

        mo_energy = mf.mo_energy
        mo_occ = mf.mo_occ
        kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]
        e_ia = numpy.concatenate( [x.reshape(-1) for x in
                                   _get_e_ia(mo_energy, mo_occ, kconserv)] )

        nov = e_ia.size
        nstates = min(nstates, nov)
        e_threshold = numpy.partition(e_ia, nstates-1)[nstates-1]
        e_threshold += self.deg_eia_thresh

        idx = numpy.where(e_ia <= e_threshold)[0]
        x0 = numpy.zeros((idx.size, nov))
        for i, j in enumerate(idx):
            x0[i, j] = 1  # Koopmans' excitations
        return x0

    def kernel(self, x0=None):
        '''TDA diagonalization solver

        Args:
            x0: list of init guess arrays for each k-shift specified in :attr:`self.kshift_lst`
                [x0_1, x0_2, ..., x0_nshift]
            x0_i ~ (nstates, nkpts*nocc*nvir)
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()

        log = logger.new_logger(self)

        mf = self._scf
        mo_occ = mf.mo_occ

        def pickeig(w, v, nroots, envs):
            idx = numpy.where(w > self.positive_eig_threshold)[0]
            return w[idx], v[:,idx], idx

        log = logger.Logger(self.stdout, self.verbose)

        self.converged = []
        self.e = []
        self.xy = []
        for i,kshift in enumerate(self.kshift_lst):
            kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

            vind, hdiag = self.gen_vind(self._scf, kshift)
            precond = self.get_precond(hdiag)

            if x0 is None:
                x0k = self.get_init_guess(self._scf, kshift, self.nstates)
            else:
                x0k = x0[i]

            converged, e, x1 = lr_eigh(
                vind, x0k, precond, tol_residual=self.conv_tol, lindep=self.lindep,
                nroots=self.nstates, pick=pickeig, max_cycle=self.max_cycle,
                max_memory=self.max_memory, verbose=log)
            self.converged.append( converged )
            self.e.append( e )
            # 1/sqrt(2) because self.x is for alpha excitation amplitude and 2(X^+*X) = 1
            self.xy.append( [(_unpack(xi*numpy.sqrt(.5), mo_occ, kconserv), 0) for xi in x1] )

        log.timer(self.__class__.__name__, *cpu0)
        self._finalize()
        return self.e, self.xy
CIS = KTDA = TDA


class TDHF(KTDBase):

    get_ab = TDA.get_ab

    def gen_vind(self, mf, kshift=0):
        '''
        [ A   B ][X]
        [-B* -A*][Y]
        '''
        assert mf is self._scf
        assert kshift == 0

        singlet = self.singlet

        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        nkpts = len(mo_occ)
        nao, nmo = mo_coeff[0].shape
        occidx = [mo_occ[k]==2 for k in range(nkpts)]
        viridx = [mo_occ[k]==0 for k in range(nkpts)]
        orbo = [mo_coeff[k][:,occidx[k]] for k in range(nkpts)]
        orbv = [mo_coeff[k][:,viridx[k]] for k in range(nkpts)]
        dtype = numpy.result_type(*mo_coeff)

        kconserv = numpy.arange(nkpts)
        e_ia = _get_e_ia(scf.addons.mo_energy_with_exxdiv_none(mf), mo_occ, kconserv)
        hdiag = numpy.hstack([x.ravel() for x in e_ia])
        tot_x = hdiag.size
        hdiag = numpy.hstack((hdiag, -hdiag))

        mem_now = lib.current_memory()[0]
        max_memory = max(2000, self.max_memory*.8-mem_now)
        vresp = self.gen_response(singlet=singlet, hermi=0, max_memory=max_memory)

        def vind(xys):
            nz = len(xys)
            z1xs = [_unpack(xy[:tot_x], mo_occ, kconserv) for xy in xys]
            z1ys = [_unpack(xy[tot_x:], mo_occ, kconserv) for xy in xys]
            dms = numpy.zeros((nz,nkpts,nao,nao), dtype=dtype)
            for i in range(nz):
                for k in range(nkpts):
                    # *2 for double occupancy
                    dmx = z1xs[i][k] * 2
                    dmy = z1ys[i][k] * 2
                    dms[i,k] += lib.einsum('ov,pv,qo->pq', dmx, orbv[k], orbo[k].conj())
                    dms[i,k] += lib.einsum('ov,qv,po->pq', dmy, orbv[k].conj(), orbo[k])

            with lib.temporary_env(mf, exxdiv=None):
                v1ao = vresp(dms, kshift) # = <mb||nj> Xjb + <mj||nb> Yjb
            v1s = []
            for i in range(nz):
                dmx = z1xs[i]
                dmy = z1ys[i]
                v1xs = [0] * nkpts
                v1ys = [0] * nkpts
                for k in range(nkpts):
                    # AX + BY
                    # = <aj||ib> Xjb + <ab||ij> Yjb
                    # = (<mj||nb> Xjb + <mb||nj> Yjb) Cma* Cni
                    v1x = lib.einsum('pq,qo,pv->ov', v1ao[i,k], orbo[k], orbv[k].conj())
                    # (B*)X + (A*)Y
                    # = <ij||ab> Xjb + <ib||aj> Yjb
                    # = (<mj||nb> Xjb + <mb||nj> Yjb) Cmi* Cna
                    v1y = lib.einsum('pq,po,qv->ov', v1ao[i,k], orbo[k].conj(), orbv[k])
                    v1x += e_ia[k] * dmx[k]
                    v1y += e_ia[k] * dmy[k]
                    v1xs[k] += v1x.ravel()
                    v1ys[k] -= v1y.ravel()
                v1s.append( numpy.concatenate(v1xs + v1ys) )
            return numpy.stack(v1s)
        return vind, hdiag

    def get_init_guess(self, mf, kshift, nstates=None):
        x0 = TDA.get_init_guess(self, mf, kshift, nstates)
        y0 = numpy.zeros_like(x0)
        return numpy.hstack([x0, y0])

    get_precond = rhf.TDHF.get_precond

    def kernel(self, x0=None):
        '''TDHF diagonalization with non-Hermitian eigenvalue solver
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()

        log = logger.new_logger(self)

        mf = self._scf
        mo_occ = mf.mo_occ

        real_system = (is_gamma_point(self._scf.kpts) and
                       self._scf.mo_coeff[0].dtype == numpy.double)

        if any(k != 0 for k in self.kshift_lst):
            # It's not clear how to define the Y matrix for kshift!=0 .
            # When the A tensor is constructed against the X(kshift) matrix,
            # the diagonal terms e_ia are calculated as e_i[k] - e_k[k+kshift].
            # Given the k-conserve relation in the A tensor, the j-b indices in
            # the A tensor should follow j[k'], b[k'+kshift]. This leads to the
            # j-b indices in the B tensor being defined as (j[k'+shift], b[k']).
            # To form the square A-B-B-A matrix, the diagonal terms for the
            # -A* part need to be constructed as e_i[k+kshift] - e_a[k], which
            # conflict to the diagonal terms of the A tensor.
            raise RuntimeError('kshift != 0 for TDHF')

        # We only need positive eigenvalues
        def pickeig(w, v, nroots, envs):
            realidx = numpy.where((abs(w.imag) < REAL_EIG_THRESHOLD) &
                                  (w.real > self.positive_eig_threshold))[0]
            return lib.linalg_helper._eigs_cmplx2real(w, v, realidx, real_system)

        log = logger.Logger(self.stdout, self.verbose)

        def norm_xy(z, kconserv):
            x, y = z.reshape(2,-1)
            norm = lib.norm(x)**2 - lib.norm(y)**2
            if norm < 0:
                log.warn('TDDFT amplitudes |X| smaller than |Y|')
            norm = abs(.5/norm)**.5
            x *= norm
            y *= norm
            return _unpack(x, mo_occ, kconserv), _unpack(y, mo_occ, kconserv)

        self.converged = []
        self.e = []
        self.xy = []
        for i,kshift in enumerate(self.kshift_lst):
            kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

            vind, hdiag = self.gen_vind(self._scf, kshift)
            precond = self.get_precond(hdiag)

            if x0 is None:
                x0k = self.get_init_guess(self._scf, kshift, self.nstates)
            else:
                x0k = x0[i]

            converged, e, x1 = lr_eig(
                vind, x0k, precond, tol_residual=self.conv_tol, lindep=self.lindep,
                nroots=self.nstates, pick=pickeig, max_cycle=self.max_cycle,
                max_memory=self.max_memory, verbose=log)
            self.converged.append( converged )
            self.e.append( e )
            self.xy.append( [norm_xy(z, kconserv) for z in x1] )

        log.timer(self.__class__.__name__, *cpu0)
        self._finalize()
        return self.e, self.xy
RPA = KTDHF = TDHF

def _get_e_ia(mo_energy, mo_occ, kconserv=None):
    nkpts = len(mo_occ)
    e_ia = [None] * nkpts
    if kconserv is None: kconserv = numpy.arange(nkpts)
    for k, kp in enumerate(kconserv):
        moeocc = mo_energy[k][mo_occ[k] > 1e-6]
        moevir = mo_energy[kp][mo_occ[kp] < 1e-6]
        e_ia[k] = -moeocc[:,None] + moevir
    return e_ia

def _unpack(vo, mo_occ, kconserv):
    z = []
    p1 = 0
    no_kpts = [numpy.count_nonzero(occ) for occ in mo_occ]
    for k, no in enumerate(no_kpts):
        kp = kconserv[k]
        nv = mo_occ[kp].size - no_kpts[kp]
        p0, p1 = p1, p1 + no * nv
        z.append(vo[p0:p1].reshape(no,nv))
    return z


scf.khf.KRHF.TDA  = lib.class_as_method(KTDA)
scf.khf.KRHF.TDHF = lib.class_as_method(KTDHF)
scf.krohf.KROHF.TDA  = NotImplemented
scf.krohf.KROHF.TDHF = NotImplemented



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Matthew Hermes
# Author: Matthew Hennefarth <mhennefarth@uchicago.com>

import numpy as np
from pyscf import __config__, lib
from pyscf.lib import logger
from pyscf.dft import numint
from pyscf.mcpdft.otpd import _grid_ao2mo
import ctypes

SWITCH_SIZE = getattr(__config__, 'dft_numint_SWITCH_SIZE', 800)
libpdft = lib.load_library('libpdft')


# TODO: outcore implementation; can I use properties instead of copying?
class _ERIS:
    '''Stores two-body PDFT on-top effective integral arrays in a form
    compatible with existing MC-SCF kernel and derivative functions.
    Unlike actual eris, PDFT 2-electron effective integrals have 24-fold
    permutation symmetry, so j_pc = k_pc and ppaa = papa.transpose
    (0,2,1,3). The mcscf _ERIS is currently undocumented so I won't
    spend more time documenting this for now.
    '''

    def __init__(self, mol, mo_coeff, ncore, ncas, method='incore',
                 paaa_only=False, aaaa_only=False, jk_pc=False, verbose=0, stdout=None):
        self.mol = mol
        self.mo_coeff = mo_coeff
        self.nao, self.nmo = mo_coeff.shape
        self.ncore = ncore
        self.ncas = ncas
        self.vhf_c = np.zeros((self.nmo, self.nmo), dtype=mo_coeff.dtype)
        self.method = method
        self.paaa_only = paaa_only
        self.aaaa_only = aaaa_only
        self.jk_pc = jk_pc
        self.verbose = verbose
        self.stdout = stdout
        if method == 'incore':
            self.papa = np.zeros((self.nmo, ncas, self.nmo, ncas),
                                 dtype=mo_coeff.dtype)
            self.j_pc = np.zeros((self.nmo, ncore), dtype=mo_coeff.dtype)
        else:
            raise NotImplementedError("method={} for pdft_eff2".format(
                self.method))

    def _accumulate(self, ot, ao, weight, rho_c, rho_a, eff_Pi,
                    non0tab=None, shls_slice=None, ao_loc=None):
        args = [ot, ao, weight, rho_c, rho_a, eff_Pi, non0tab, shls_slice, ao_loc]
        self._accumulate_vhf_c(*args)
        if self.method.lower() == 'incore':
            self._accumulate_ppaa_incore(*args)
        else:
            raise NotImplementedError("method={} for pdft_eff2".format(
                self.method))
        self._accumulate_j_pc(*args)

    def _accumulate_vhf_c(self, ot, ao, weight, rho_c, rho_a, eff_Pi,
                          non0tab, shls_slice, ao_loc):
        mo_coeff = self.mo_coeff
        ncore, ncas = self.ncore, self.ncas
        nocc = ncore + ncas

        vrho_c = _contract_eff_rho(eff_Pi, rho_c)
        self.vhf_c += mo_coeff.conjugate().T @ ot.get_eff_1body(ao,
                                                             weight, vrho_c, non0tab=non0tab, shls_slice=shls_slice,
                                                             ao_loc=ao_loc,
                                                             hermi=1) @ mo_coeff
        self.energy_core = np.trace(self.vhf_c[:ncore, :ncore]) #/ 2
        if self.paaa_only:
            # 1/2 v_aiuv D_ii D_uv = v^ai_uv D_uv -> F_ai, F_ia
            # needs to be in here since it would otherwise be calculated using
            # ppaa and papa. This is harmless to the CI problem because the
            # elements in the active space and core-core sector are ignored
            # below.
            eff_rho_a = _contract_eff_rho(eff_Pi, rho_a)
            vhf_a = get_eff_1body(ot, ao, weight, eff_rho_a, non0tab=non0tab,
                                  shls_slice=shls_slice, ao_loc=ao_loc, hermi=1)
            vhf_a = mo_coeff.conjugate().T @ vhf_a @ mo_coeff
            vhf_a[ncore:nocc, :] = vhf_a[:, ncore:nocc] = 0.0
            self.vhf_c += vhf_a

    def _ftpt_vhf_c(self):
        return self.nao + 1

    def _accumulate_ppaa_incore(self, ot, ao, weight, rho_c, rho_a,
                                eff_Pi, non0tab, shls_slice, ao_loc):
        # ao is here stored in row-major order = deriv,AOs,grids regardless of
        # what the ndarray object thinks
        mo_coeff = self.mo_coeff
        ncore, ncas = self.ncore, self.ncas
        nocc = ncore + ncas
        nderiv = eff_Pi.shape[0]
        mo_cas = _grid_ao2mo(self.mol, ao[:nderiv], mo_coeff[:, ncore:nocc],
                             non0tab)
        if self.aaaa_only:
            aaaa = ot.get_eff_2body([mo_cas, mo_cas, mo_cas,
                                      mo_cas], weight, eff_Pi, aosym='s1')
            self.papa[ncore:nocc, :, ncore:nocc, :] += aaaa
        elif self.paaa_only:
            paaa = ot.get_eff_2body([ao, mo_cas, mo_cas, mo_cas],
                                 weight, eff_Pi, aosym='s1')
            paaa = np.tensordot(mo_coeff.T, paaa, axes=1)
            self.papa[:, :, ncore:nocc, :] += paaa
            self.papa[ncore:nocc, :, :, :] += paaa.transpose(2, 3, 0, 1)
            self.papa[ncore:nocc, :, ncore:nocc, :] -= paaa[ncore:nocc, :, :, :]
        else:
            papa = ot.get_eff_2body([ao, mo_cas, ao, mo_cas],
                                 weight, eff_Pi, aosym='s1')
            papa = np.tensordot(mo_coeff.T, papa, axes=1)
            self.papa += np.tensordot(mo_coeff.T, papa,
                                      axes=((1), (2))).transpose(1, 2, 0, 3)

    def _ftpt_ppaa_incore(self):
        nao, ncas = self.nao, self.ncas
        ncol = 1 + 2 * ncas
        ij_aa = int(self.aaaa_only)
        kl_aa = int(ij_aa or self.paaa_only)
        ncol += (ij_aa + kl_aa) * (nao - ncas)
        return ncol * ncas

    def _accumulate_j_pc(self, ot, ao, weight, rho_c, rho_a, vPi,
                         non0tab, shls_slice, ao_loc):
        mo_coeff = self.mo_coeff
        ncore = self.ncore
        nderiv = vPi.shape[0]
        if self.jk_pc:
            mo = _square_ao(_grid_ao2mo(self.mol, ao[:nderiv], mo_coeff,
                                        non0tab))
            mo_core = mo[:, :, :ncore]
            self.j_pc += ot.get_eff_1body([mo, mo_core], weight, vPi)

    def _ftpt_j_pc(self):
        return self.nao + self.ncore + self.ncas + 1

    def _accumulate_ftpt(self):
        """ memory footprint of _accumulate, divided by nderiv_Pi*ngrids """
        ftpt_fns = [self._ftpt_vhf_c]
        if self.method.lower() == 'incore':
            ftpt_fns.append(self._ftpt_ppaa_incore)
        else:
            raise NotImplementedError("method={} for pdft_eff2".format(
                self.method))
        if self.verbose > logger.DEBUG:
            ftpt_fns.append(self._ftpt_j_pc)
        ncol = 0
        for fn in ftpt_fns: ncol = max(ncol, fn())
        return ncol

    def _finalize(self):
        if self.method == 'incore':
            self.ppaa = np.ascontiguousarray(self.papa.transpose(0, 2, 1, 3))
            self.k_pc = self.j_pc.copy()
        else:
            raise NotImplementedError("method={} for pdft_eff2".format(
                self.method))
        self.k_pc = self.j_pc.copy()


def _contract_eff_rho(eff, rho, add_eff_rho=None):
    """ Make a jk-like eff_rho from eff and a density. k = j so it's just
        eff * eff_rho / 2 , but the product rule needs to be followed """
    if rho.ndim == 1:
        rho = rho[None, :]

    nderiv = eff.shape[0]
    eff_rho = eff * rho[0]

    if nderiv > 1:
        eff_rho[0] += (eff[1:4] * rho[1:4]).sum(0)

    eff_rho /= 2
    # eff involves lower derivatives than eff_rho in original translation
    # make sure vot * rho gets added to only the proper component(s)
    if add_eff_rho is not None:
        add_eff_rho[:nderiv] += eff_rho
        eff_rho = add_eff_rho
    return eff_rho


def _square_ao(ao):
    # On a grid, square each element of an AO or MO array, but preserve the
    # chain rule so that columns 1 to 4 are still the first derivative of
    # the squared AO value, etc.
    nderiv = ao.shape[0]
    ao_sq = ao * ao[0]
    if nderiv > 1:
        ao_sq[1:4] *= 2
    if nderiv > 4:
        ao_sq[4:10] += ao[1:4] ** 2
        ao_sq[4:10] *= 2
    return ao_sq

def get_eff_1body(otfnal, ao, weight, kern, non0tab=None,
                  shls_slice=None, ao_loc=None, hermi=0):
    r''' Contract the kern with d vrho/ dDpq.

    Args:
        ao : ndarray or 2 ndarrays of shape (*,ngrids,nao)
            contains values and derivatives of nao.
            2 different ndarrays can have different nao but not
            different ngrids
        weight : ndarray of shape (ngrids)
            containing numerical integration weights
        kern : ndarray of shape (*,ngrids)
            the derivative of the on-top potential with respect to
            density (vrho)/ If not provided, it is calculated.

    Kwargs:
        non0tab : ndarray of shape (nblk, nbas)
            Identifies blocks of grid points which are nonzero on
            each AO shell so as to exploit sparsity.
            If you want the "ao" array to be in the MO basis, just
            leave this as None. If hermi == 0, it only applies
            to the bra index ao array, even if the ket index ao
            array is the same (so probably always pass hermi = 1
            in that case)
        shls_slice : sequence of integers of len 2
            Identifies starting and stopping indices of AO shells
        ao_loc : ndarray of length nbas
            Offset to first AO of each shell
        hermi : integer or logical
            Toggle whether veff is supposed to be a Hermitian matrix
            You can still pass two different ao arrays for the bra and
            the ket indices, for instance if one of them is supposed to
            be a higher derivative. They just have to have the same nao
            in that case.

    Returns : ndarray of shape (nao[0],nao[1])
        The 1-body effective term corresponding to kernel times the AO's,
        in the atomic-orbital basis. In PDFT this functional is always
        spin-symmetric.
    '''
    if isinstance(ao, np.ndarray) and ao.ndim == 3:
        ao = [ao, ao]
    elif len(ao) != 2:
        raise NotImplementedError("uninterpretable aos!")
    elif ao[0].size < ao[1].size:
        # Life pro-tip: do more operations with smaller arrays and fewer
        # operations with bigger arrays
        ao = [ao[1], ao[0]]

    kern = kern.copy()
    kern *= weight[None, :]

    # Zeroth and first derivatives
    nderiv = kern.shape[0]

    first_pass = min(nderiv, 4)
    eff_ao = _contract_kern_ao(kern[:first_pass], ao[1][:first_pass])
    nterm = eff_ao.shape[0]
    eff = sum([_dot_ao_mo(otfnal.mol, a, v, non0tab=non0tab,
                          shls_slice=shls_slice, ao_loc=ao_loc, hermi=hermi)
               for a, v in zip(ao[0][0:nterm], eff_ao)])

    if nderiv > 4:
        # check if we have laplacian...
        vtau = kern[4]

        vtau *= 0.5
        eff += _tau_dot_ao_mo(
            otfnal.mol,
            ao[1],
            ao[0],
            vtau,
            non0tab=non0tab,
            shls_slice=shls_slice,
            ao_loc=ao_loc,
            hermi=hermi,
        )

    if nderiv > 5:
        raise NotImplementedError("laplacian translated meta-GGA functional")

    return eff

def get_eff_2body(otfnal, ao, weight, kern, aosym='s4', eff_ao=None):
    if isinstance(ao, np.ndarray) and ao.ndim == 3:
        ao = [ao, ao, ao, ao]
    elif len(ao) != 4:
        raise NotImplementedError('fancy orbital subsets and fast evaluation '
                                  'in get_eff_2body')

    if isinstance(aosym, int):
        aosym = str(aosym)
    ij_symm = "4" in aosym or "2ij" in aosym
    kl_symm = "4" in aosym or "2kl" in aosym

    if eff_ao is None:
        eff_ao = otfnal.get_eff_2body_kl(ao[2], ao[3], weight, kern, symm=kl_symm)

    nderiv = eff_ao.shape[0]
    ao2 = _contract_ao1_ao2(ao[0], ao[1], nderiv, symm=ij_symm)
    try:
        ao2 = ao2.transpose(0, 3, 2, 1)

    except ValueError as e:
        print(ao[0].shape, ao[1].shape, ao2.shape)
        raise(e)

    eff_ao = eff_ao.transpose(0, 3, 2, 1)
    ijkl_shape = list(ao2.shape[1:-1]) + list(eff_ao.shape[1:-1])
    ao2 = ao2.reshape(ao2.shape[0], -1, ao2.shape[-1]).transpose(0, 2, 1)
    eff_ao = eff_ao.reshape(eff_ao.shape[0], -1, eff_ao.shape[-1]).transpose(0, 2, 1)
    eff = sum([_dot_ao_mo(otfnal.mol, a, v) for a, v in zip(ao2, eff_ao)])
    eff = eff.reshape(*ijkl_shape)

    return eff


def get_eff_2body_kl(ot, ao_k, ao_l, weight, kern, symm=False):
    kern = kern.copy()
    kern *= weight[None, :]

    # Flatten deriv and grid so I can tensordot it all at once
    # Index symmetry can be built into _contract_ao1_ao2
    kern_ao = _contract_kern_ao(kern, ao_l)
    eff_ao = _contract_ao_eff_ao(ao_k, kern_ao, symm=symm)
    return eff_ao

def _dot_ao_mo(mol, ao, mo, non0tab=None, shls_slice=None, ao_loc=None,
               hermi=0):
    # f_ij = int g_i(r) * h_j(r) dr
    # Like numint._dot_ao_ao, but allows for two different bases on the
    # rows and columns
    if hermi:
        return numint._dot_ao_ao(mol, ao, mo, non0tab=non0tab,
                                 shls_slice=shls_slice, ao_loc=ao_loc, hermi=hermi)
    ngrids, nao = ao.shape
    nmo = mo.shape[-1]

    if nao < SWITCH_SIZE:
        return lib.dot(ao.T.conj(), mo)

    if not ao.flags.f_contiguous:
        ao = lib.transpose(ao)
    if not mo.flags.f_contiguous:
        mo = lib.transpose(mo)
    if ao.dtype == mo.dtype == np.double:
        fn = libpdft.VOTdot_ao_mo
    else:
        raise NotImplementedError("Complex-orbital PDFT")

    if non0tab is None or shls_slice is None or ao_loc is None:
        pnon0tab = pshls_slice = pao_loc = lib.c_null_ptr()
    else:
        pnon0tab = non0tab.ctypes.data_as(ctypes.c_void_p)
        pshls_slice = (ctypes.c_int * 2)(*shls_slice)
        pao_loc = ao_loc.ctypes.data_as(ctypes.c_void_p)

    vv = np.empty((nao, nmo), dtype=ao.dtype)
    fn(vv.ctypes.data_as(ctypes.c_void_p),
       ao.ctypes.data_as(ctypes.c_void_p),
       mo.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(nao), ctypes.c_int(nmo),
       ctypes.c_int(ngrids), ctypes.c_int(mol.nbas),
       pnon0tab, pshls_slice, pao_loc)
    return vv


# TODO: unittest?
def _contract_kern_ao(vot, ao, out=None):
    # Evaluate v_i(r) = v(r) * AO_i(r) and its derivatives on a grid
    # Note that the chain rule means that v' * AO' -> v, v' * AO -> v'
    ''' REQUIRES array in shape = (nderiv,nao,ngrids) and data layout
        = (nderiv,ngrids,nao)/row-major '''
    nderiv = vot.shape[0]
    ao = np.ascontiguousarray(ao.transpose(0, 2, 1))
    nao, ngrids = ao.shape[1:]
    vao = np.ndarray((nderiv, nao, ngrids), dtype=ao.dtype,
                     buffer=out).transpose(0, 2, 1)
    ao = ao.transpose(0, 2, 1)
    vao[0] = numint._scale_ao(ao[:nderiv], vot, out=vao[0])
    if nderiv > 1:
        for i in range(1, 4):
            vao[i] = numint._scale_ao(ao[0:1, :, :], vot[i:i + 1, :], out=vao[i])

    return vao


def _contract_ao_eff_ao(ao, vao, symm=False):
    r''' Outer-product of ao grid and eff * ao grid
    Can be used with two-orb-dimensional vao if the last two dimensions
    are flattened into "nao"

    Args:
        ao : ndarray of shape (*,ngrids,nao1)
        vao : ndarray of shape (nderiv,ngrids,nao2)

    Kwargs:
        symm : logical
            If true, nao1 == nao2 must be true

    Returns: ndarray of shape (nderiv,ngrids,nao1,nao2)
        or (nderiv,ngrids,nao1*(nao1+1)//2)
    '''
    ao = ao.transpose(0, 2, 1)
    vao = vao.transpose(0, 2, 1)
    assert ao.flags.c_contiguous, 'shape = {} ; strides = {}'.format(
        ao.shape, ao.strides)
    assert vao.flags.c_contiguous, 'shape = {} ; strides = {}'.format(
        vao.shape, vao.strides)

    nderiv = vao.shape[0]
    if symm:
        ix_p, ix_q = np.tril_indices(ao.shape[1])
        ao = ao[:, ix_p]
        vao = vao[:, ix_q]
    else:
        ao = np.expand_dims(ao, -2)
        vao = np.expand_dims(vao, -3)
    prod = ao[0] * vao
    if nderiv > 1:
        prod[0] += (ao[1:4] * vao[1:4]).sum(0)
    if symm:
        prod = prod.transpose(0, 2, 1)
    else:
        prod = prod.transpose(0, 3, 2, 1)
    return prod


# TODO: unittest?
def _contract_ao1_ao2(ao1, ao2, nderiv, symm=False):
    # Evaluate P_ij(r) = AO_i(r) * AO_j(r) and its derivatives on a grid
    ao1 = ao1.transpose(0, 2, 1)
    ao2 = ao2.transpose(0, 2, 1)
    assert (ao1.flags.c_contiguous), 'shape = {} ; strides = {}'.format(
        ao1.shape, ao1.strides)
    assert (ao2.flags.c_contiguous), 'shape = {} ; strides = {}'.format(
        ao2.shape, ao2.strides)
    if symm:  # TODO: C implementation of this slow indexing
        ix_p, ix_q = np.tril_indices(ao1.shape[1])
        ao1 = ao1[:nderiv, ix_p]
        ao2 = ao2[:nderiv, ix_q]
    else:
        ao1 = np.expand_dims(ao1, -2)[:nderiv]
        ao2 = np.expand_dims(ao2, -3)[:nderiv]
    prod = ao1[:nderiv] * ao2[0]
    if nderiv > 1:
        prod[1:4] += ao1[0] * ao2[1:4]  # Product rule
    ao2 = None
    if symm:
        prod = prod.transpose(0, 2, 1)
    else:
        prod = prod.transpose(0, 3, 2, 1)
    return prod


def _tau_dot_ao_mo(
    mol, ao, mo, vtau, non0tab=None, shls_slice=None, ao_loc=None, hermi=0
):
    vao = numint._scale_ao(ao[1], vtau)
    eff = _dot_ao_mo(
        mol,
        mo[1],
        vao,
        non0tab=non0tab,
        shls_slice=shls_slice,
        ao_loc=ao_loc,
        hermi=hermi,
    )
    vao = numint._scale_ao(ao[2], vtau)
    eff += _dot_ao_mo(
        mol,
        mo[2],
        vao,
        non0tab=non0tab,
        shls_slice=shls_slice,
        ao_loc=ao_loc,
        hermi=hermi,
    )
    vao = numint._scale_ao(ao[3], vtau)
    eff += _dot_ao_mo(
        mol,
        mo[3],
        vao,
        non0tab=non0tab,
        shls_slice=shls_slice,
        ao_loc=ao_loc,
        hermi=hermi,
    )
    return eff



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from pyscf import lib
from pyscf.pbc import dft
from pyscf.tdscf import uks
from pyscf.pbc.tdscf.uhf import TDA
from pyscf.pbc.tdscf.uhf import TDHF as TDDFT

RPA = TDUKS = TDDFT


class CasidaTDDFT(TDDFT):
    get_init_guess = TDA.get_init_guess
    _gen_vind = uks.TDDFTNoHybrid.gen_vind
    gen_vind = TDA.gen_vind
    kernel = uks.TDDFTNoHybrid.kernel

TDDFTNoHybrid = CasidaTDDFT

def tddft(mf, frozen=None):
    '''Driver to create TDDFT or CasidaTDDFT object'''
    if mf._numint.libxc.is_hybrid_xc(mf.xc):
        return TDDFT(mf, frozen)
    else:
        return CasidaTDDFT(mf, frozen)

dft.uks.UKS.TDA           = lib.class_as_method(TDA)
dft.uks.UKS.TDHF          = None
dft.uks.UKS.TDDFTNoHybrid = lib.class_as_method(TDDFTNoHybrid)
dft.uks.UKS.CasidaTDDFT   = lib.class_as_method(CasidaTDDFT)
dft.uks.UKS.TDDFT         = tddft
#dft.rks.RKS.dTDA          = lib.class_as_method(dTDA)
#dft.rks.RKS.dRPA          = lib.class_as_method(dRPA)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from itertools import product
from scipy import linalg
from pyscf import gto, dft, ao2mo, fci, mcscf, lib
from pyscf.lib import logger, temporary_env
from pyscf.mcscf.addons import StateAverageMCSCFSolver, state_average_mix, state_average_mix_, state_average
from pyscf.fci import direct_spin1
from pyscf import mcpdft

def coulomb_tensor (mc, mo_coeff=None, ci=None, h2eff=None, eris=None):
    '''Compute w_IJKL = (tu|vx) D^IJ_tu D^KL_vx

    Args:
        mc : mcscf method instance

    Kwargs:
        mo_coeff : ndarray of shape (nao,nmo)
            Contains molecular orbital coefficients
        ci : list of ndarrays of shape (ndeta,ndetb)
            Contains CI vectors
        h2eff : ndarray of shape [ncas,]*4
            Contains active-space ERIs
        eris : mc_ao2mo.ERI object
            Contains active-space ERIs. Ignored if h2eff is passed; if
            h2eff is not passed then it is constructed from eris.ppaa

    Returns:
        w : ndarray of shape [nroots,]*4
    '''
    if mo_coeff is None: mo_coeff=mc.mo_coeff
    if ci is None: ci = mc.ci
    # TODO: state-average mix extension
    ci = np.asarray (ci)
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    nroots, nocc = mc.fcisolver.nroots, ncore + ncas
    if h2eff is None:
        if eris is None: h2eff = mc.get_h2eff (mo_coeff=mo_coeff)
        else: h2eff = np.asarray (eris.ppaa[ncore:nocc,ncore:nocc,:,:])
    h2eff = ao2mo.restore (1, h2eff, ncas)

    row, col = np.tril_indices (nroots)
    tdm1 = np.stack (mc.fcisolver.states_trans_rdm12(ci[col], ci[row], ncas,
        nelecas)[0], axis=0)

    w = np.tensordot (tdm1, h2eff, axes=2)
    w = np.tensordot (w, tdm1, axes=((1,2),(1,2)))
    return ao2mo.restore (1, w, nroots)

def e_coul (mc, mo_coeff=None, ci=None, h2eff=None, eris=None):
    '''Compute the sum of active-space Coulomb energies (the diabatizer
    function for CMS-PDFT) and its first and second derivatives

    Args:
        mc : mcscf method instance

    Kwargs:
        mo_coeff : ndarray of shape (nao,nmo)
            Contains molecular orbital coefficients
        ci : list of ndarrays of shape (ndeta,ndetb)
            Contains CI vectors
        h2eff : ndarray of shape [ncas,]*4
            Contains active-space ERIs
        eris : mc_ao2mo.ERI object
            Contains active-space ERIs. Ignored if h2eff is passed; if
            h2eff is not passed then it is constructed from eris.ppaa

    Returns:
        Qaa : float
            sum of Coulomb energies
        dQaa : ndarray of shape npair = nroots*(nroots-1)/2
            first derivatives of J wrt interstate rotation
        d2Qaa : ndarray of shape (npair,npair)
            Hessian of J wrt interstate rotation
        Qaa_update : callable
            Takes a unitary matrix of shape (nroots, nroots) and returns
            Qaa, dQaa, and d2Qaa as above using the stored Coulomb
            tensor intermediate from this function.
    '''
    nroots = mc.fcisolver.nroots

    w0 = coulomb_tensor (mc, mo_coeff=mo_coeff, ci=ci, h2eff=h2eff,
        eris=eris)
    Qaa0, dQaa0, d2Qaa0 = _e_coul (w0, nroots)
    def Qaa_update (u=1):
        w1 = ao2mo.incore.full (w0, u, compact=False)
        return _e_coul (w1, nroots)
    return Qaa0, dQaa0, d2Qaa0, Qaa_update

def _e_coul (w_IJKL, nroots):
    npair = nroots * (nroots - 1) // 2
    w_IJKK = np.diagonal (w_IJKL, axis1=2, axis2=3)
    w_IKJK = np.diagonal (w_IJKL, axis1=1, axis2=3)
    w_IJJJ = np.diagonal (w_IJKK, axis1=1, axis2=2)

    Qaa = np.trace (w_IJJJ) / 2.0

    tril_mask = np.zeros ([nroots,nroots], dtype=np.bool_)
    tril_mask[np.tril_indices (nroots,k=-1)] = True
    dQaa = 2*(w_IJJJ.T-w_IJJJ)[tril_mask]
    # My sign convention is row idx = source state; col idx = dest
    # state, lower-triangular positive. The Newton iteration is designed
    # with this in mind and breaks if I flip it. However, regardless of
    # sign convention, the unitary operator parameterized this way
    # always comes out with destination on the rows and source on the
    # columns, because that's just what the word "operator" means:
    # |f> = U|i>. So when I exponentiate later, I transpose. Don't let
    # the fact that that transpose is mathematically the same as
    # flipping this sign confuse you: the sign here is CORRECT.

    v_IJ_K = -4*w_IKJK - 2*w_IJKK
    v_IJ_K += (w_IJJJ+w_IJJJ.T)[:,:,None]
    d2Qaa = np.zeros_like (w_IJKL)
    for k in range (nroots):
        d2Qaa[:,k,k,:] = v_IJ_K[:,:,k]
    d2Qaa -= d2Qaa.transpose (0,1,3,2)
    d2Qaa -= d2Qaa.transpose (1,0,2,3)
    tril_mask2 = np.logical_and.outer (tril_mask, tril_mask)
    d2Qaa = d2Qaa[tril_mask2].reshape (npair, npair)

    return Qaa, dQaa, d2Qaa

def e_coul_o0 (mc,ci):
    # Old implementation
    nroots = mc.fcisolver.nroots
    ncas, ncore = mc.ncas,mc.ncore
    nocc = ncas + ncore
    rows,col = np.tril_indices(nroots,k=-1)
    pairs = len(rows)
    mo_cas = mc.mo_coeff[:,ncore:nocc]
    ci_array = np.array(ci)
    casdm1 = mc.fcisolver.states_make_rdm1 (ci,ncas,mc.nelecas)
    dm1 = np.dot(casdm1,mo_cas.T)
    dm1 = np.dot(mo_cas,dm1).transpose(1,0,2)
    j = mc._scf.get_j (dm=dm1)
    e_coul = (j*dm1).sum((1,2)) / 2

    trans12_tdm1, trans12_tdm2 = mc.fcisolver.states_trans_rdm12(ci_array[col],
        ci_array[rows],ncas,mc.nelecas)
    trans12_tdm1_array = np.array(trans12_tdm1)
    tdm1 = np.dot(trans12_tdm1_array,mo_cas.T)
    tdm1 = np.dot(mo_cas,tdm1).transpose(1,0,2)
    rowscol2ind = np.zeros ((nroots, nroots), dtype=int)
    rowscol2ind[(rows,col)] = list (range (pairs))
    rowscol2ind += rowscol2ind.T
    rowscol2ind[np.diag_indices(nroots)] = -1

    def w_klmn(k,l,m,n,dm,tdm):
        d = dm[k] if k==l else tdm[rowscol2ind[k,l]]
        dm1_g = mc._scf.get_j (dm=d)
        d = dm[m] if m==n else tdm[rowscol2ind[m,n]]
        w = (dm1_g*d).sum ((0,1))
        return w

    def v_klmn(k,l,m,n,dm,tdm):
        if l==m:
            v = (w_klmn(k,n,k,k,dm,tdm)-w_klmn(k,n,l,l,dm,tdm)
                +w_klmn(n,k,n,n,dm,tdm)-w_klmn(k,n,m,m,dm,tdm)
                -4*w_klmn(k,l,m,n,dm,tdm))
        else:
            v = 0
        return v

    dg = mc._scf.get_j (dm=tdm1)
    grad1 = (dg*dm1[rows]).sum((1,2))
    grad2 = (dg*dm1[col]).sum((1,2))
    e_grad = np.zeros(pairs)
    e_grad = 2*(grad1 - grad2)

    e_hess = np.zeros((pairs,pairs))
    for (i, (k,l)), (j, (m,n)) in product (enumerate (zip (rows, col)),
            repeat=2):
        e_hess[i,j] = (v_klmn(k,l,m,n,dm1,tdm1)+v_klmn(l,k,n,m,dm1,tdm1)
                      -v_klmn(k,l,n,m,dm1,tdm1)-v_klmn(l,k,m,n,dm1,tdm1))

    return sum (e_coul), e_grad, e_hess




# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
NIST physical constants

https://physics.nist.gov/cuu/Constants/
https://physics.nist.gov/cuu/Constants/Table/allascii.txt
'''

LIGHT_SPEED = 137.03599967994   # http://physics.nist.gov/cgi-bin/cuu/Value?alph
# BOHR = .529 177 210 92(17) e-10m  # http://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
BOHR = 0.52917721092  # Angstroms
BOHR_SI = BOHR * 1e-10

ALPHA = 7.2973525664e-3         # http://physics.nist.gov/cgi-bin/cuu/Value?alph
G_ELECTRON = 2.00231930436182   # http://physics.nist.gov/cgi-bin/cuu/Value?gem
E_MASS = 9.10938356e-31         # kg https://physics.nist.gov/cgi-bin/cuu/Value?me
AVOGADRO = 6.022140857e23       # https://physics.nist.gov/cgi-bin/cuu/Value?na
ATOMIC_MASS = 1e-3/AVOGADRO
PROTON_MASS = 1.672621898e-27   # kg https://physics.nist.gov/cgi-bin/cuu/Value?mp
PROTON_MASS_AU = PROTON_MASS/ATOMIC_MASS
MP_ME = PROTON_MASS / E_MASS    # Proton-to-electron mass ratio
BOHR_MAGNETON = 927.4009994e-26 # J/T http://physics.nist.gov/cgi-bin/cuu/Value?mub
NUC_MAGNETON = BOHR_MAGNETON / MP_ME
PLANCK = 6.626070040e-34        # J*s http://physics.nist.gov/cgi-bin/cuu/Value?h
HBAR = PLANCK/(2*3.141592653589793) # https://physics.nist.gov/cgi-bin/cuu/Value?hbar
#HARTREE2J = 4.359744650e-18     # J https://physics.nist.gov/cgi-bin/cuu/Value?hrj
HARTREE2J = HBAR**2/(E_MASS*BOHR_SI**2)
HARTREE2EV = 27.21138602        # eV https://physics.nist.gov/cgi-bin/cuu/Value?threv
E_CHARGE = 1.6021766208e-19     # C https://physics.nist.gov/cgi-bin/cuu/Value?e
LIGHT_SPEED_SI = 299792458      # https://physics.nist.gov/cgi-bin/cuu/Value?c
DEBYE = 3.335641e-30            # C*m = 1e-18/LIGHT_SPEED_SI https://cccbdb.nist.gov/debye.asp
AU2DEBYE = E_CHARGE * BOHR*1e-10 / DEBYE # 2.541746
AUEFG = 9.71736235660e21        # V/m^2 https://physics.nist.gov/cgi-bin/cuu/Value?auefg
AU2TESLA = HBAR/(BOHR_SI**2 * E_CHARGE)
BOLTZMANN = 1.38064852e-23      # J/K https://physics.nist.gov/cgi-bin/cuu/Value?k
HARTREE2WAVENUMBER = 1e-2 * HARTREE2J / (LIGHT_SPEED_SI * PLANCK) # 2.194746313702e5
AMU2AU = ATOMIC_MASS/E_MASS



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Full CI solver for spin-free Hamiltonian.  This solver can be used to compute
doublet, triplet,...

The CI wfn are stored as a 2D array [alpha,beta], where each row corresponds
to an alpha string.  For each row (alpha string), there are
total-num-beta-strings of columns.  Each column corresponds to a beta string.

Different FCI solvers are implemented to support different type of symmetry.
                    Symmetry
File                Point group   Spin singlet   Real hermitian*    Alpha/beta degeneracy
direct_spin0_symm   Yes           Yes            Yes                Yes
direct_spin1_symm   Yes           No             Yes                Yes
direct_spin0        No            Yes            Yes                Yes
direct_spin1        No            No             Yes                Yes
direct_uhf          No            No             Yes                No
direct_nosym        No            No             No**               Yes

*  Real hermitian Hamiltonian implies (ij|kl) = (ji|kl) = (ij|lk) = (ji|lk)
** Hamiltonian is real but not hermitian, (ij|kl) != (ji|kl) ...
'''

import sys
import ctypes
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import ao2mo
from pyscf.lib import logger
from pyscf.fci import cistring
from pyscf.fci import rdm
from pyscf.fci import spin_op
from pyscf.fci import addons
from pyscf.fci.spin_op import contract_ss
from pyscf.fci.addons import _unpack_nelec, civec_spinless_repr
from pyscf import __config__

libfci = cistring.libfci

def contract_1e(f1e, fcivec, norb, nelec, link_index=None):
    '''Contract the 1-electron Hamiltonian with a FCI vector to get a new FCI
    vector.
    '''
    fcivec = numpy.asarray(fcivec, order='C')
    link_indexa, link_indexb = _unpack(norb, nelec, link_index)
    na, nlinka = link_indexa.shape[:2]
    nb, nlinkb = link_indexb.shape[:2]
    assert fcivec.size == na*nb
    assert fcivec.dtype == f1e.dtype == numpy.float64
    f1e_tril = lib.pack_tril(f1e)
    ci1 = numpy.zeros_like(fcivec)
    libfci.FCIcontract_a_1e(f1e_tril.ctypes.data_as(ctypes.c_void_p),
                            fcivec.ctypes.data_as(ctypes.c_void_p),
                            ci1.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb),
                            ctypes.c_int(na), ctypes.c_int(nb),
                            ctypes.c_int(nlinka), ctypes.c_int(nlinkb),
                            link_indexa.ctypes.data_as(ctypes.c_void_p),
                            link_indexb.ctypes.data_as(ctypes.c_void_p))
    libfci.FCIcontract_b_1e(f1e_tril.ctypes.data_as(ctypes.c_void_p),
                            fcivec.ctypes.data_as(ctypes.c_void_p),
                            ci1.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb),
                            ctypes.c_int(na), ctypes.c_int(nb),
                            ctypes.c_int(nlinka), ctypes.c_int(nlinkb),
                            link_indexa.ctypes.data_as(ctypes.c_void_p),
                            link_indexb.ctypes.data_as(ctypes.c_void_p))
    return ci1.view(FCIvector)

def contract_2e(eri, fcivec, norb, nelec, link_index=None):
    r'''Contract the 4-index tensor eri[pqrs] with a FCI vector

    .. math::

        |output\rangle = E_{pq} E_{rs} eri_{pq,rs} |CI\rangle \\

        E_{pq}E_{rs} = E_{pr,qs} + \delta_{qr} E_{ps} \\

        E_{pq} = p^+ q + \bar{p}^+ \bar{q}

        E_{pr,qs} = p^+ r^+ s q + \bar{p}^+ r^+ s \bar{q} + ...

    :math:`p,q,...` means spin-up orbitals and :math:`\bar{p}, \bar{q}` means
    spin-down orbitals.

    Note the input argument eri is NOT the 2e hamiltonian tensor. 2e hamiltonian is

    .. math::

        h2e &= (pq|rs) E_{pr,qs} \\
            &= (pq|rs) (E_{pq}E_{rs} - \delta_{qr} E_{ps}) \\
            &= eri_{pq,rs} E_{pq}E_{rs} \\

    So the relation between eri and hamiltonian (the 2e-integral tensor) is

    .. math::

        eri_{pq,rs} = (pq|rs) - (1/Nelec) \sum_q (pq|qs)

    to restore the symmetry between pq and rs,

    .. math::

        eri_{pq,rs} = (pq|rs) - (.5/Nelec) [\sum_q (pq|qs) + \sum_p (pq|rp)]

    See also :func:`direct_spin1.absorb_h1e`
    '''
    fcivec = numpy.asarray(fcivec, order='C')
    eri = numpy.asarray(ao2mo.restore(4, eri, norb), order='C')
    link_indexa, link_indexb = _unpack(norb, nelec, link_index)
    na, nlinka = link_indexa.shape[:2]
    nb, nlinkb = link_indexb.shape[:2]
    assert fcivec.size == na*nb
    assert fcivec.dtype == eri.dtype == numpy.float64
    ci1 = numpy.empty_like(fcivec)

    libfci.FCIcontract_2e_spin1(eri.ctypes.data_as(ctypes.c_void_p),
                                fcivec.ctypes.data_as(ctypes.c_void_p),
                                ci1.ctypes.data_as(ctypes.c_void_p),
                                ctypes.c_int(norb),
                                ctypes.c_int(na), ctypes.c_int(nb),
                                ctypes.c_int(nlinka), ctypes.c_int(nlinkb),
                                link_indexa.ctypes.data_as(ctypes.c_void_p),
                                link_indexb.ctypes.data_as(ctypes.c_void_p))
    return ci1.view(FCIvector)

def make_hdiag(h1e, eri, norb, nelec, compress=False):
    '''Diagonal Hamiltonian for Davidson preconditioner

    Kwargs:
        compress (bool) : whether to remove symmetry forbidden elements
    '''
    if not (h1e.dtype == eri.dtype == numpy.float64):
        raise NotImplementedError('Complex Hamiltonian')

    neleca, nelecb = _unpack_nelec(nelec)
    h1e = numpy.asarray(h1e, order='C')
    eri = numpy.asarray(ao2mo.restore(1, eri, norb), order='C')
    occslsta = occslstb = cistring.gen_occslst(range(norb), neleca)
    if neleca != nelecb:
        occslstb = cistring.gen_occslst(range(norb), nelecb)
    na = len(occslsta)
    nb = len(occslstb)

    hdiag = numpy.empty(na*nb)
    jdiag = numpy.asarray(numpy.einsum('iijj->ij',eri), order='C')
    kdiag = numpy.asarray(numpy.einsum('ijji->ij',eri), order='C')
    c_h1e = h1e.ctypes.data_as(ctypes.c_void_p)
    c_jdiag = jdiag.ctypes.data_as(ctypes.c_void_p)
    c_kdiag = kdiag.ctypes.data_as(ctypes.c_void_p)
    libfci.FCImake_hdiag_uhf(hdiag.ctypes.data_as(ctypes.c_void_p),
                             c_h1e, c_h1e, c_jdiag, c_jdiag, c_jdiag, c_kdiag, c_kdiag,
                             ctypes.c_int(norb),
                             ctypes.c_int(na), ctypes.c_int(nb),
                             ctypes.c_int(neleca), ctypes.c_int(nelecb),
                             occslsta.ctypes.data_as(ctypes.c_void_p),
                             occslstb.ctypes.data_as(ctypes.c_void_p))
    return hdiag

def absorb_h1e(h1e, eri, norb, nelec, fac=1):
    '''Modify 2e Hamiltonian to include 1e Hamiltonian contribution.
    '''
    if h1e.dtype == numpy.complex128 or eri.dtype == numpy.complex128:
        raise NotImplementedError('Complex Hamiltonian')

    if not isinstance(nelec, (int, numpy.number)):
        nelec = sum(nelec)
    h2e = ao2mo.restore(1, eri.copy(), norb)
    f1e = h1e - numpy.einsum('jiik->jk', h2e) * .5
    f1e = f1e * (1./(nelec+1e-100))
    for k in range(norb):
        h2e[k,k,:,:] += f1e
        h2e[:,:,k,k] += f1e
    return ao2mo.restore(4, h2e, norb) * fac

def pspace(h1e, eri, norb, nelec, hdiag=None, np=400):
    '''pspace Hamiltonian to improve Davidson preconditioner. See, CPL, 169, 463
    '''
    if norb >= 64:
        raise NotImplementedError('norb >= 64')

    if h1e.dtype == numpy.complex128 or eri.dtype == numpy.complex128:
        raise NotImplementedError('Complex Hamiltonian')

    neleca, nelecb = _unpack_nelec(nelec)
    h1e = numpy.ascontiguousarray(h1e)
    eri = ao2mo.restore(1, eri, norb)
    na = cistring.num_strings(norb, neleca)
    nb = cistring.num_strings(norb, nelecb)
    if hdiag is None:
        hdiag = make_hdiag(h1e, eri, norb, nelec, compress=False)
    assert hdiag.size == na * nb
    if hdiag.size <= np:
        addr = numpy.arange(hdiag.size)
    else:
        try:
            addr = numpy.argpartition(hdiag, np-1)[:np].copy()
        except AttributeError:
            addr = numpy.argsort(hdiag)[:np].copy()
    addra, addrb = divmod(addr, nb)
    stra = cistring.addrs2str(norb, neleca, addra)
    strb = cistring.addrs2str(norb, nelecb, addrb)
    np = len(addr)
    h0 = numpy.zeros((np,np))
    libfci.FCIpspace_h0tril(h0.ctypes.data_as(ctypes.c_void_p),
                            h1e.ctypes.data_as(ctypes.c_void_p),
                            eri.ctypes.data_as(ctypes.c_void_p),
                            stra.ctypes.data_as(ctypes.c_void_p),
                            strb.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_int(norb), ctypes.c_int(np))

    HERMITIAN_THRESHOLD = 1e-10
    if (abs(h1e - h1e.T).max() < HERMITIAN_THRESHOLD and
        abs(eri - eri.transpose(1,0,3,2)).max() < HERMITIAN_THRESHOLD):
        # symmetric Hamiltonian
        h0 = lib.hermi_triu(h0)
    else:
        # Fill the upper triangular part
        h0 = numpy.asarray(h0, order='F')
        h1e = numpy.asarray(h1e.T, order='C')
        eri = numpy.asarray(eri.transpose(1,0,3,2), order='C')
        libfci.FCIpspace_h0tril(h0.ctypes.data_as(ctypes.c_void_p),
                                h1e.ctypes.data_as(ctypes.c_void_p),
                                eri.ctypes.data_as(ctypes.c_void_p),
                                stra.ctypes.data_as(ctypes.c_void_p),
                                strb.ctypes.data_as(ctypes.c_void_p),
                                ctypes.c_int(norb), ctypes.c_int(np))

    idx = numpy.arange(np)
    h0[idx,idx] = hdiag[addr]
    return addr, h0

# be careful with single determinant initial guess. It may diverge the
# preconditioner when the eigvalue of first davidson iter equals to hdiag
def kernel(h1e, eri, norb, nelec, ci0=None, level_shift=1e-3, tol=1e-10,
           lindep=1e-14, max_cycle=50, max_space=12, nroots=1,
           davidson_only=False, pspace_size=400, orbsym=None, wfnsym=None,
           ecore=0, **kwargs):
    return _kfactory(FCISolver, h1e, eri, norb, nelec, ci0, level_shift,
                     tol, lindep, max_cycle, max_space, nroots,
                     davidson_only, pspace_size, ecore=ecore, **kwargs)
def _kfactory(Solver, h1e, eri, norb, nelec, ci0=None, level_shift=1e-3,
              tol=1e-10, lindep=1e-14, max_cycle=50, max_space=12, nroots=1,
              davidson_only=False, pspace_size=400, ecore=0, **kwargs):
    cis = Solver(None)
    cis.level_shift = level_shift
    cis.conv_tol = tol
    cis.lindep = lindep
    cis.max_cycle = max_cycle
    cis.max_space = max_space
    cis.nroots = nroots
    cis.davidson_only = davidson_only
    cis.pspace_size = pspace_size

    unknown = {}
    for k in kwargs:
        if not hasattr(cis, k):
            unknown[k] = kwargs[k]
        setattr(cis, k, kwargs[k])
    if unknown:
        sys.stderr.write('Unknown keys %s for FCI kernel %s\n' %
                         (str(unknown.keys()), __name__))
    e, c = cis.kernel(h1e, eri, norb, nelec, ci0, ecore=ecore, **unknown)
    return e, c

def energy(h1e, eri, fcivec, norb, nelec, link_index=None):
    '''Compute the FCI electronic energy for given Hamiltonian and FCI vector.
    '''
    h2e = absorb_h1e(h1e, eri, norb, nelec, .5)
    ci1 = contract_2e(h2e, fcivec, norb, nelec, link_index)
    return numpy.dot(fcivec.reshape(-1), ci1.reshape(-1))


def make_rdm1s(fcivec, norb, nelec, link_index=None):
    r'''Spin separated 1-particle density matrices.
    The return values include two density matrices: (alpha,alpha), (beta,beta)

    dm1[p,q] = <q^\dagger p>

    The convention is based on McWeeney's book, Eq (5.4.20).
    The contraction between 1-particle Hamiltonian and rdm1 is
    E = einsum('pq,qp', h1, rdm1)
    '''
    if link_index is None:
        neleca, nelecb = _unpack_nelec(nelec)
        link_indexa = cistring.gen_linkstr_index(range(norb), neleca)
        link_indexb = cistring.gen_linkstr_index(range(norb), nelecb)
        link_index = (link_indexa, link_indexb)
    rdm1a = rdm.make_rdm1_spin1('FCImake_rdm1a', fcivec, fcivec,
                                norb, nelec, link_index)
    rdm1b = rdm.make_rdm1_spin1('FCImake_rdm1b', fcivec, fcivec,
                                norb, nelec, link_index)
    return rdm1a, rdm1b

def make_rdm1(fcivec, norb, nelec, link_index=None):
    r'''Spin-traced one-particle density matrix

    dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>

    The convention is based on McWeeney's book, Eq (5.4.20)
    The contraction between 1-particle Hamiltonian and rdm1 is
    E = einsum('pq,qp', h1, rdm1)
    '''
    rdm1a, rdm1b = make_rdm1s(fcivec, norb, nelec, link_index)
    return rdm1a + rdm1b

def make_rdm12s(fcivec, norb, nelec, link_index=None, reorder=True):
    r'''Spin separated 1- and 2-particle density matrices.
    The return values include two lists, a list of 1-particle density matrices
    and a list of 2-particle density matrices.  The density matrices are:
    (alpha,alpha), (beta,beta) for 1-particle density matrices;
    (alpha,alpha,alpha,alpha), (alpha,alpha,beta,beta),
    (beta,beta,beta,beta) for 2-particle density matrices.

    1pdm[p,q] = :math:`\langle q^\dagger p\rangle`;
    2pdm[p,q,r,s] = :math:`\langle p^\dagger r^\dagger s q\rangle`.

    Energy should be computed as
    E = einsum('pq,qp', h1, 1pdm) + 1/2 * einsum('pqrs,pqrs', eri, 2pdm)
    where h1[p,q] = <p|h|q> and eri[p,q,r,s] = (pq|rs)
    '''
    dm1a, dm2aa = rdm.make_rdm12_spin1('FCIrdm12kern_a', fcivec, fcivec,
                                       norb, nelec, link_index, 1)
    dm1b, dm2bb = rdm.make_rdm12_spin1('FCIrdm12kern_b', fcivec, fcivec,
                                       norb, nelec, link_index, 1)
    _, dm2ab = rdm.make_rdm12_spin1('FCItdm12kern_ab', fcivec, fcivec,
                                    norb, nelec, link_index, 0)
    if reorder:
        dm1a, dm2aa = rdm.reorder_rdm(dm1a, dm2aa, inplace=True)
        dm1b, dm2bb = rdm.reorder_rdm(dm1b, dm2bb, inplace=True)
    return (dm1a, dm1b), (dm2aa, dm2ab, dm2bb)

def make_rdm12(fcivec, norb, nelec, link_index=None, reorder=True):
    r'''Spin traced 1- and 2-particle density matrices.

    1pdm[p,q] = :math:`\langle q_\alpha^\dagger p_\alpha \rangle +
                       \langle q_\beta^\dagger  p_\beta \rangle`;
    2pdm[p,q,r,s] = :math:`\langle p_\alpha^\dagger r_\alpha^\dagger s_\alpha q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\alpha^\dagger s_\alpha q_\beta\rangle +
                           \langle p_\alpha^\dagger r_\beta^\dagger  s_\beta  q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\beta^\dagger  s_\beta  q_\beta\rangle`.

    Energy should be computed as
    E = einsum('pq,qp', h1, 1pdm) + 1/2 * einsum('pqrs,pqrs', eri, 2pdm)
    where h1[p,q] = <p|h|q> and eri[p,q,r,s] = (pq|rs)
    '''
    #(dm1a, dm1b), (dm2aa, dm2ab, dm2bb) = \
    #        make_rdm12s(fcivec, norb, nelec, link_index, reorder)
    #return dm1a+dm1b, dm2aa+dm2ab+dm2ab.transpose(2,3,0,1)+dm2bb
    dm1, dm2 = rdm.make_rdm12_spin1('FCIrdm12kern_sf', fcivec, fcivec,
                                    norb, nelec, link_index, 1)
    if reorder:
        dm1, dm2 = rdm.reorder_rdm(dm1, dm2, inplace=True)
    return dm1, dm2

def make_rdm123(fcivec, norb, nelec, link_index=None, reorder=True):
    '''Spin traced 1-, 2-, and 3-particle density matrices.'''
    dm1, dm2, dm3 = rdm.make_dm123('FCI3pdm_kern_sf', fcivec, fcivec, norb, nelec)
    if reorder:
        dm1, dm2, dm3 = rdm.reorder_dm123(dm1, dm2, dm3, inplace=True)
    return dm1, dm2, dm3

def make_rdm123s(fcivec, norb, nelec, link_index=None, reorder=True):
    r'''Spin separated 1-, 2-, and 3-particle density matrices.

    1pdm[p,q] = :math:`\langle q_\alpha^\dagger p_\alpha \rangle +
                       \langle q_\beta^\dagger  p_\beta \rangle`;
    2pdm[p,q,r,s] = :math:`\langle p_\alpha^\dagger r_\alpha^\dagger s_\alpha q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\alpha^\dagger s_\alpha q_\beta\rangle +
                           \langle p_\alpha^\dagger r_\beta^\dagger  s_\beta  q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\beta^\dagger  s_\beta  q_\beta\rangle`.
    '''
    if (not reorder):
        raise NotImplementedError('reorder=False not currently supported')
    ci_spinless = civec_spinless_repr([fcivec,], norb, [nelec,])
    rdm1, rdm2, rdm3 = make_rdm123(ci_spinless, norb*2, (nelec[0]+nelec[1],0))

    rdm1a = rdm1[:norb,:norb]
    rdm1b = rdm1[norb:,norb:]
    # assert np.allclose(rdm1a+rdm1b, rdm1)

    rdm2aa = rdm2[:norb,:norb,:norb,:norb]
    rdm2ab = rdm2[:norb,:norb,norb:,norb:]
    rdm2bb = rdm2[norb:,norb:,norb:,norb:]
    # assert np.allclose(rdm2aa+rdm2bb+rdm2ab+rdm2ab.transpose(2,3,0,1), rdm2)

    rdm3aaa = rdm3[:norb,:norb,:norb,:norb,:norb,:norb]
    rdm3aab = rdm3[:norb,:norb,:norb,:norb,norb:,norb:]
    rdm3abb = rdm3[:norb,:norb,norb:,norb:,norb:,norb:]
    rdm3bbb = rdm3[norb:,norb:,norb:,norb:,norb:,norb:]
    # assert np.allclose(rdm3aaa+rdm3bbb+rdm3aab+rdm3aab.transpose(0,1,4,5,2,3)+\
    # rdm3aab.transpose(4,5,0,1,2,3)+rdm3abb+rdm3abb.transpose(2,3,0,1,4,5)+rdm3abb.transpose(2,3,4,5,0,1), rdm3)
    return (rdm1a, rdm1b), (rdm2aa, rdm2ab, rdm2bb), (rdm3aaa, rdm3aab, rdm3abb, rdm3bbb)

def make_rdm1234(fcivec, norb, nelec, link_index=None, reorder=True):
    '''Spin traced 1-, 2-, 3, 4-particle density matrices.'''
    dm1, dm2, dm3, dm4 = rdm.make_dm1234('FCI4pdm_kern_sf', fcivec, fcivec, norb, nelec)
    if reorder:
        dm1, dm2, dm3, dm4 = rdm.reorder_dm1234(dm1, dm2, dm3, dm4, inplace=True)
    return dm1, dm2, dm3, dm4

def make_rdm1234s(fcivec, norb, nelec, link_index=None, reorder=True):
    r'''Spin separated 1-, 2-, 3, 4-particle density matrices.

    1pdm[p,q] = :math:`\langle q_\alpha^\dagger p_\alpha \rangle +
                       \langle q_\beta^\dagger  p_\beta \rangle`;
    2pdm[p,q,r,s] = :math:`\langle p_\alpha^\dagger r_\alpha^\dagger s_\alpha q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\alpha^\dagger s_\alpha q_\beta\rangle +
                           \langle p_\alpha^\dagger r_\beta^\dagger  s_\beta  q_\alpha\rangle +
                           \langle p_\beta^\dagger  r_\beta^\dagger  s_\beta  q_\beta\rangle`.
    '''
    if (not reorder):
        raise NotImplementedError('reorder=False not currently supported')
    ci_spinless = civec_spinless_repr([fcivec,], norb, [nelec,])
    rdm1, rdm2, rdm3, rdm4 = make_rdm1234(ci_spinless, norb*2, (nelec[0]+nelec[1],0))

    # define slices
    alpha = slice(0, norb)
    beta = slice(norb, None)

    rdm1a = rdm1[alpha, alpha]
    rdm1b = rdm1[beta, beta]
    # assert np.allclose(rdm1a+rdm1b, rdm1)

    rdm2aa = rdm2[alpha, alpha, alpha, alpha]
    rdm2ab = rdm2[alpha, alpha, beta, beta]
    rdm2bb = rdm2[beta, beta, beta, beta]
    # assert np.allclose(rdm2aa+rdm2bb+rdm2ab+rdm2ab.transpose(2,3,0,1), rdm2)

    rdm3aaa = rdm3[alpha, alpha, alpha, alpha, alpha, alpha]
    rdm3aab = rdm3[alpha, alpha, alpha, alpha, beta, beta]
    rdm3abb = rdm3[alpha, alpha, beta, beta, beta, beta]
    rdm3bbb = rdm3[beta, beta, beta, beta, beta, beta]
    # assert np.allclose(rdm3aaa+rdm3bbb+rdm3aab+rdm3aab.transpose(0,1,4,5,2,3)+\
    # rdm3aab.transpose(4,5,0,1,2,3)+rdm3abb+rdm3abb.transpose(2,3,0,1,4,5)+rdm3abb.transpose(2,3,4,5,0,1), rdm3)

    rdm4aaaa = rdm4[alpha, alpha, alpha, alpha, alpha, alpha, alpha, alpha]
    rdm4aaab = rdm4[alpha, alpha, alpha, alpha, alpha, alpha, beta, beta]
    rdm4aabb = rdm4[alpha, alpha, alpha, alpha, beta, beta, beta, beta]
    rdm4abbb = rdm4[alpha, alpha, beta, beta, beta, beta, beta, beta]
    rdm4bbbb = rdm4[beta, beta, beta, beta, beta, beta, beta, beta]
    # assert numpy.allclose(
    #     rdm4aaaa
    #     + rdm4bbbb
    #     + rdm4aaab
    #     + rdm4aaab.transpose(0, 1, 2, 3, 6, 7, 4, 5)
    #     + rdm4aaab.transpose(0, 1, 6, 7, 2, 3, 4, 5)
    #     + rdm4aaab.transpose(6, 7, 0, 1, 2, 3, 4, 5)
    #     + rdm4aabb
    #     + rdm4aabb.transpose(0, 1, 4, 5, 2, 3, 6, 7)
    #     + rdm4aabb.transpose(0, 1, 4, 5, 6, 7, 2, 3)
    #     + rdm4aabb.transpose(4, 5, 2, 3, 0, 1, 6, 7)
    #     + rdm4aabb.transpose(4, 5, 2, 3, 6, 7, 0, 1)
    #     + rdm4abbb
    #     + rdm4abbb.transpose(2, 3, 0, 1, 4, 5, 6, 7)
    #     + rdm4abbb.transpose(2, 3, 4, 5, 0, 1, 6, 7)
    #     + rdm4abbb.transpose(2, 3, 4, 5, 6, 7, 0, 1),
    #     rdm4,
    # )
    return (
        (rdm1a, rdm1b),
        (rdm2aa, rdm2ab, rdm2bb),
        (rdm3aaa, rdm3aab, rdm3abb, rdm3bbb),
        (rdm4aaaa, rdm4aaab, rdm4aabb, rdm4abbb, rdm4bbbb),
    )

def trans_rdm1s(cibra, ciket, norb, nelec, link_index=None):
    r'''Spin separated transition 1-particle density matrices.
    The return values include two density matrices: (alpha,alpha), (beta,beta).
    See also function :func:`make_rdm1s`

    1pdm[p,q] = :math:`\langle q^\dagger p \rangle`
    '''
    rdm1a = rdm.make_rdm1_spin1('FCItrans_rdm1a', cibra, ciket,
                                norb, nelec, link_index)
    rdm1b = rdm.make_rdm1_spin1('FCItrans_rdm1b', cibra, ciket,
                                norb, nelec, link_index)
    return rdm1a, rdm1b

def trans_rdm1(cibra, ciket, norb, nelec, link_index=None):
    r'''Spin traced transition 1-particle transition density matrices.

    1pdm[p,q] = :math:`\langle q_\alpha^\dagger p_\alpha \rangle
                       + \langle q_\beta^\dagger p_\beta \rangle`
    '''
    rdm1a, rdm1b = trans_rdm1s(cibra, ciket, norb, nelec, link_index)
    return rdm1a + rdm1b

def trans_rdm12s(cibra, ciket, norb, nelec, link_index=None, reorder=True):
    r'''Spin separated 1- and 2-particle transition density matrices.
    The return values include two lists, a list of 1-particle transition
    density matrices and a list of 2-particle transition density matrices.
    The density matrices are:
    (alpha,alpha), (beta,beta) for 1-particle transition density matrices;
    (alpha,alpha,alpha,alpha), (alpha,alpha,beta,beta),
    (beta,beta,alpha,alpha), (beta,beta,beta,beta) for 2-particle transition
    density matrices.

    1pdm[p,q] = :math:`\langle q^\dagger p\rangle`;
    2pdm[p,q,r,s] = :math:`\langle p^\dagger r^\dagger s q\rangle`.
    '''
    dm1a, dm2aa = rdm.make_rdm12_spin1('FCItdm12kern_a', cibra, ciket,
                                       norb, nelec, link_index, 2)
    dm1b, dm2bb = rdm.make_rdm12_spin1('FCItdm12kern_b', cibra, ciket,
                                       norb, nelec, link_index, 2)
    _, dm2ab = rdm.make_rdm12_spin1('FCItdm12kern_ab', cibra, ciket,
                                    norb, nelec, link_index, 0)
    _, dm2ba = rdm.make_rdm12_spin1('FCItdm12kern_ab', ciket, cibra,
                                    norb, nelec, link_index, 0)
    dm2ba = dm2ba.transpose(3,2,1,0)
    if reorder:
        dm1a, dm2aa = rdm.reorder_rdm(dm1a, dm2aa, inplace=True)
        dm1b, dm2bb = rdm.reorder_rdm(dm1b, dm2bb, inplace=True)
    return (dm1a, dm1b), (dm2aa, dm2ab, dm2ba, dm2bb)

def trans_rdm12(cibra, ciket, norb, nelec, link_index=None, reorder=True):
    r'''Spin traced transition 1- and 2-particle transition density matrices.

    1pdm[p,q] = :math:`\langle q^\dagger p\rangle`;
    2pdm[p,q,r,s] = :math:`\langle p^\dagger r^\dagger s q\rangle`.
    '''
    #(dm1a, dm1b), (dm2aa, dm2ab, dm2ba, dm2bb) = \
    #        trans_rdm12s(cibra, ciket, norb, nelec, link_index, reorder)
    #return dm1a+dm1b, dm2aa+dm2ab+dm2ba+dm2bb
    dm1, dm2 = rdm.make_rdm12_spin1('FCItdm12kern_sf', cibra, ciket,
                                    norb, nelec, link_index, 2)
    if reorder:
        dm1, dm2 = rdm.reorder_rdm(dm1, dm2, inplace=True)
    return dm1, dm2

def _get_init_guess(na, nb, nroots, hdiag, nelec):
    # The "nroots" lowest determinats based on energy expectation value.
    ci0 = []
    neleca, nelecb = _unpack_nelec(nelec)
    if neleca == nelecb and na == nb:
        hdiag = lib.pack_tril(hdiag.reshape(na, na))
        if hdiag.size <= nroots:
            addrs = numpy.arange(hdiag.size)
        else:
            addrs = numpy.argpartition(hdiag, nroots-1)[:nroots]
        for addr in addrs:
            addra = (int)((2*addr+.25)**.5 - .5 + 1e-7)
            addrb = addr - addra*(addra+1)//2
            x = numpy.zeros((na, na))
            x[addra,addrb] = 1
            ci0.append(x.ravel().view(FCIvector))
    else:
        if hdiag.size <= nroots:
            addrs = numpy.arange(hdiag.size)
        else:
            addrs = numpy.argpartition(hdiag, nroots-1)[:nroots]
        for addr in addrs:
            x = numpy.zeros((na*nb))
            x[addr] = 1
            ci0.append(x.view(FCIvector))

    # Add noise
    ci0[0][0 ] += 1e-5
    ci0[0][-1] -= 1e-5
    return ci0

def get_init_guess(norb, nelec, nroots, hdiag):
    '''Initial guess is the single Slater determinant
    '''
    neleca, nelecb = _unpack_nelec(nelec)
    na = cistring.num_strings(norb, neleca)
    nb = cistring.num_strings(norb, nelecb)
    return _get_init_guess(na, nb, nroots, hdiag, nelec)


###############################################################
# direct-CI driver
###############################################################

def kernel_ms1(fci, h1e, eri, norb, nelec, ci0=None, link_index=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None, hop=None,
               max_memory=None, verbose=None, ecore=0, **kwargs):
    '''
    Args:
        h1e: ndarray
            1-electron Hamiltonian
        eri: ndarray
            2-electron integrals in chemist's notation
        norb: int
            Number of orbitals
        nelec: int or (int, int)
            Number of electrons of the system

    Kwargs:
        ci0: ndarray
            Initial guess
        link_index: ndarray
            A lookup table to cache the addresses of CI determinants in
            wave-function vector
        tol: float
            Convergence tolerance
        lindep: float
            Linear dependence threshold
        max_cycle: int
            Max. iterations for diagonalization
        max_space: int
            Max. trial vectors to store for sub-space diagonalization method
        nroots: int
            Number of states to solve
        davidson_only: bool
            Whether to call subspace diagonalization (davidson solver) or do a
            full diagonalization (lapack eigh) for small systems
        pspace_size: int
            Number of determinants as the threshold of "small systems",
        hop: function(c) => array_like_c
            Function to use for the Hamiltonian multiplication with trial vector

    Note: davidson solver requires more arguments. For the parameters not
    dispatched, they can be passed to davidson solver via the extra keyword
    arguments **kwargs
    '''
    if nroots is None: nroots = fci.nroots
    if davidson_only is None: davidson_only = fci.davidson_only
    if pspace_size is None: pspace_size = fci.pspace_size
    if max_memory is None:
        max_memory = fci.max_memory - lib.current_memory()[0]
    log = logger.new_logger(fci, verbose)

    nelec = _unpack_nelec(nelec, fci.spin)
    assert (0 <= nelec[0] <= norb and 0 <= nelec[1] <= norb)

    hdiag = fci.make_hdiag(h1e, eri, norb, nelec, compress=False).ravel()
    num_dets = hdiag.size
    pspace_size = min(num_dets, pspace_size)
    addr = [0]
    pw = pv = None
    if pspace_size > 0 and norb < 64:
        addr, h0 = fci.pspace(h1e, eri, norb, nelec, hdiag, pspace_size)
        pw, pv = fci.eig(h0)
        pspace_size = len(addr)

    if getattr(fci, 'sym_allowed_idx', None):
        # Remove symmetry forbidden elements
        sym_idx = numpy.hstack(fci.sym_allowed_idx)
        civec_size = sym_idx.size
    else:
        sym_idx = None
        civec_size = num_dets

    if max_memory < civec_size*6*8e-6:
        log.warn('Not enough memory for FCI solver. '
                 'The minimal requirement is %.0f MB', civec_size*60e-6)

    if pspace_size >= civec_size and ci0 is None and not davidson_only:
        if nroots > 1:
            nroots = min(civec_size, nroots)
            civec = numpy.empty((nroots,civec_size))
            civec[:,addr] = pv[:,:nroots].T
            return pw[:nroots]+ecore, civec
        elif pspace_size == 1 or abs(pw[0]-pw[1]) > 1e-12:
            # Check degeneracy. Degenerated wfn may break point group symmetry.
            # Davidson iteration with a proper initial guess can avoid this problem.
            civec = numpy.empty(civec_size)
            civec[addr] = pv[:,0]
            return pw[0]+ecore, civec
    pw = pv = h0 = None

    if sym_idx is None:
        precond = fci.make_precond(hdiag)
    else:
        precond = fci.make_precond(hdiag[sym_idx])

    h2e = fci.absorb_h1e(h1e, eri, norb, nelec, .5)
    if hop is None:
        cpu0 = [logger.process_clock(), logger.perf_counter()]
        def hop(c):
            hc = fci.contract_2e(h2e, c, norb, nelec, link_index)
            cpu0[:] = log.timer_debug1('contract_2e', *cpu0)
            return hc.ravel()

    def init_guess():
        if callable(getattr(fci, 'get_init_guess', None)):
            return fci.get_init_guess(norb, nelec, nroots, hdiag)
        else:
            x0 = []
            for i in range(min(len(addr), nroots)):
                x = numpy.zeros(civec_size)
                x[addr[i]] = 1
                x0.append(x)
            return x0

    if ci0 is None:
        ci0 = init_guess  # lazy initialization to reduce memory footprint
    elif not callable(ci0):
        if isinstance(ci0, numpy.ndarray):
            ci0 = [ci0.ravel()]
        else:
            ci0 = [x.ravel() for x in ci0]
        if sym_idx is not None and ci0[0].size != civec_size:
            ci0 = [x[sym_idx] for x in ci0]
        # If provided initial guess ci0 are accidentally the eigenvectors of the
        # system, Davidson solver may be failed to find enough roots as it is
        # unable to generate more subspace basis from ci0. Adding vectors so
        # initial guess to help Davidson solver generate enough basis.
        if len(ci0) < nroots:
            ci0.extend(init_guess()[len(ci0):])

    if tol is None: tol = fci.conv_tol
    if lindep is None: lindep = fci.lindep
    if max_cycle is None: max_cycle = fci.max_cycle
    if max_space is None: max_space = fci.max_space
    tol_residual = getattr(fci, 'conv_tol_residual', None)

    with lib.with_omp_threads(fci.threads):
        e, c = fci.eig(hop, ci0, precond, tol=tol, lindep=lindep,
                       max_cycle=max_cycle, max_space=max_space, nroots=nroots,
                       max_memory=max_memory, verbose=log, follow_state=True,
                       tol_residual=tol_residual, **kwargs)
    return e+ecore, c

def make_pspace_precond(hdiag, pspaceig, pspaceci, addr, level_shift=0):
    # precondition with pspace Hamiltonian, CPL, 169, 463
    def precond(r, e0, x0, *args):
        #h0e0 = h0 - numpy.eye(len(addr))*(e0-level_shift)
        h0e0inv = numpy.dot(pspaceci/(pspaceig-(e0-level_shift)), pspaceci.T)
        hdiaginv = 1/(hdiag - (e0-level_shift))
        hdiaginv[abs(hdiaginv)>1e8] = 1e8
        h0x0 = x0 * hdiaginv
        #h0x0[addr] = numpy.linalg.solve(h0e0, x0[addr])
        h0x0[addr] = numpy.dot(h0e0inv, x0[addr])
        h0r = r * hdiaginv
        #h0r[addr] = numpy.linalg.solve(h0e0, r[addr])
        h0r[addr] = numpy.dot(h0e0inv, r[addr])
        e1 = numpy.dot(x0, h0r) / numpy.dot(x0, h0x0)
        x1 = r - e1*x0
        #pspace_x1 = x1[addr].copy()
        x1 *= hdiaginv
        # pspace (h0-e0)^{-1} cause diverging?
        #x1[addr] = numpy.linalg.solve(h0e0, pspace_x1)
        return x1
    return precond

def make_diag_precond(hdiag, pspaceig, pspaceci, addr, level_shift=0):
    return lib.make_diag_precond(hdiag, level_shift)


class FCIBase(lib.StreamObject):
    '''Full CI solver

    Attributes:
        verbose : int
            Print level.  Default value equals to :class:`Mole.verbose`.
        max_cycle : int
            Total number of iterations. Default is 100
        max_space : tuple of int
            Davidson iteration space size. Default is 14.
        conv_tol : float
            Energy convergence tolerance. Default is 1e-10.
        level_shift : float
            Level shift applied in the preconditioner to avoid singularity.
            Default is 1e-3
        davidson_only : bool
            By default, the entire Hamiltonian matrix will be constructed and
            diagonalized if the system is small (see attribute pspace_size).
            Setting this parameter to True will enforce the eigenvalue
            problems being solved by Davidson subspace algorithm.  This flag
            should be enabled when initial guess is given or particular spin
            symmetry or point-group symmetry is required because the initial
            guess or symmetry are completely ignored in the direct diagonalization.
        pspace_size : int
            The dimension of Hamiltonian matrix over which Davidson iteration
            algorithm will be used for the eigenvalue problem.  Default is 400.
            This is roughly corresponding to a (6e,6o) system.
        nroots : int
            Number of states to be solved.  Default is 1, the ground state.
        spin : int or None
            Spin (2S = nalpha-nbeta) of the system.  If this attribute is None,
            spin will be determined by the argument nelec (number of electrons)
            of the kernel function.
        wfnsym : str or int
            Symmetry of wavefunction.  It is used only in direct_spin1_symm
            and direct_spin0_symm solver.

    Saved results

        eci : float or a list of float
            FCI energy(ies)
        ci : nparray
            FCI wfn vector(s)
        converged : bool (or a list of bool for multiple roots)
            Whether davidson iteration is converged

    Examples:

    >>> from pyscf import gto, scf, ao2mo, fci
    >>> mol = gto.M(atom='Li 0 0 0; Li 0 0 1', basis='sto-3g')
    >>> mf = scf.RHF(mol).run()
    >>> h1 = mf.mo_coeff.T.dot(mf.get_hcore()).dot(mf.mo_coeff)
    >>> eri = ao2mo.kernel(mol, mf.mo_coeff)
    >>> cisolver = fci.direct_spin1.FCI(mol)
    >>> e, ci = cisolver.kernel(h1, eri, h1.shape[1], mol.nelec, ecore=mol.energy_nuc())
    >>> print(e)
    -14.4197890826
    '''

    max_cycle = getattr(__config__, 'fci_direct_spin1_FCI_max_cycle', 100)
    max_space = getattr(__config__, 'fci_direct_spin1_FCI_max_space', 12)
    conv_tol = getattr(__config__, 'fci_direct_spin1_FCI_conv_tol', 1e-10)
    conv_tol_residual = getattr(__config__, 'fci_direct_spin1_FCI_conv_tol_residual', None)
    lindep = getattr(__config__, 'fci_direct_spin1_FCI_lindep', 1e-14)

    # level shift in preconditioner is helpful to avoid singularity and linear
    # dependence basis in davidson diagonalization solver
    level_shift = getattr(__config__, 'fci_direct_spin1_FCI_level_shift', 1e-3)

    # force the diagonalization use davidson iteration.  When the CI space
    # is small, the solver exactly diagonalizes the Hamiltonian.  But this
    # solution will ignore the initial guess.  Setting davidson_only can
    # enforce the solution on the initial guess state
    davidson_only = getattr(__config__, 'fci_direct_spin1_FCI_davidson_only', False)

    pspace_size = getattr(__config__, 'fci_direct_spin1_FCI_pspace_size', 400)
    threads = getattr(__config__, 'fci_direct_spin1_FCI_threads', None)
    lessio = getattr(__config__, 'fci_direct_spin1_FCI_lessio', False)

    _keys = {
        'max_cycle', 'max_space', 'conv_tol', 'lindep',
        'level_shift', 'davidson_only', 'pspace_size', 'threads', 'lessio',
        'mol', 'nroots', 'spin', 'orbsym', 'wfnsym', 'converged', 'norb',
        'nelec', 'eci', 'ci',
    }

    def __init__(self, mol=None):
        if mol is None:
            self.verbose = logger.NOTE
            self.max_memory = lib.param.MAX_MEMORY
        else:
            self.stdout = mol.stdout
            self.verbose = mol.verbose
            self.max_memory = mol.max_memory
        self.mol = mol
        self.nroots = 1
        self.spin = None
        # Initialize symmetry attributes for the compatibility with direct_spin1_symm
        # solver.  They are not used by direct_spin1 solver.
        self.orbsym = None
        self.wfnsym = None

        self.converged = False
        self.norb = None
        self.nelec = None
        self.eci = None
        self.ci = None

    @property
    def e_tot(self):
        return self.eci

    @property
    def nstates(self):
        return self.nroots
    @nstates.setter
    def nstates(self, x):
        self.nroots = x

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('******** %s ********', self.__class__)
        log.info('max. cycles = %d', self.max_cycle)
        log.info('conv_tol = %g', self.conv_tol)
        log.info('davidson only = %s', self.davidson_only)
        log.info('linear dependence = %g', self.lindep)
        log.info('level shift = %g', self.level_shift)
        log.info('max iter space = %d', self.max_space)
        log.info('max_memory %d MB', self.max_memory)
        log.info('nroots = %d', self.nroots)
        log.info('pspace_size = %d', self.pspace_size)
        log.info('spin = %s', self.spin)
        return self

    @lib.with_doc(absorb_h1e.__doc__)
    def absorb_h1e(self, h1e, eri, norb, nelec, fac=1):
        nelec = _unpack_nelec(nelec, self.spin)
        return absorb_h1e(h1e, eri, norb, nelec, fac)

    @lib.with_doc(make_hdiag.__doc__)
    def make_hdiag(self, h1e, eri, norb, nelec, compress=False):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_hdiag(h1e, eri, norb, nelec, compress)

    @lib.with_doc(pspace.__doc__)
    def pspace(self, h1e, eri, norb, nelec, hdiag=None, np=400):
        nelec = _unpack_nelec(nelec, self.spin)
        return pspace(h1e, eri, norb, nelec, hdiag, np)

    @lib.with_doc(contract_1e.__doc__)
    def contract_1e(self, f1e, fcivec, norb, nelec, link_index=None, **kwargs):
        nelec = _unpack_nelec(nelec, self.spin)
        return contract_1e(f1e, fcivec, norb, nelec, link_index, **kwargs)

    @lib.with_doc(contract_2e.__doc__)
    def contract_2e(self, eri, fcivec, norb, nelec, link_index=None, **kwargs):
        nelec = _unpack_nelec(nelec, self.spin)
        return contract_2e(eri, fcivec, norb, nelec, link_index, **kwargs)

    def eig(self, op, x0=None, precond=None, **kwargs):
        if isinstance(op, numpy.ndarray):
            self.converged = True
            return scipy.linalg.eigh(op)

        self.converged, e, ci = \
                lib.davidson1(lambda xs: [op(x) for x in xs],
                              x0, precond, lessio=self.lessio, **kwargs)
        if kwargs['nroots'] == 1:
            self.converged = self.converged[0]
            e = e[0]
            ci = ci[0]
        return e, ci

    def make_precond(self, hdiag, pspaceig=None, pspaceci=None, addr=None):
        if pspaceig is None:
            return make_diag_precond(hdiag, pspaceig, pspaceci, addr,
                                     self.level_shift)
        else:
            # Note: H0 in pspace may break symmetry.
            return make_pspace_precond(hdiag, pspaceig, pspaceci, addr,
                                       self.level_shift)

    @lib.with_doc(get_init_guess.__doc__)
    def get_init_guess(self, norb, nelec, nroots, hdiag):
        return get_init_guess(norb, nelec, nroots, hdiag)

    @lib.with_doc(kernel_ms1.__doc__)
    def kernel(self, h1e, eri, norb, nelec, ci0=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               orbsym=None, wfnsym=None, ecore=0, **kwargs):
        if nroots is None: nroots = self.nroots
        if self.verbose >= logger.WARN:
            self.check_sanity()
        self.norb = norb
        self.nelec = nelec = _unpack_nelec(nelec, self.spin)
        link_index = _unpack(norb, nelec, None)
        e, c = kernel_ms1(self, h1e, eri, norb, nelec, ci0, link_index,
                          tol, lindep, max_cycle, max_space, nroots,
                          davidson_only, pspace_size, ecore=ecore, **kwargs)
        self.eci = e

        na = link_index[0].shape[0]
        nb = link_index[1].shape[0]
        if nroots > 1:
            self.ci = [x.reshape(na,nb).view(FCIvector) for x in c]
        else:
            self.ci = c.reshape(na,nb).view(FCIvector)
        return self.eci, self.ci

    @lib.with_doc(energy.__doc__)
    def energy(self, h1e, eri, fcivec, norb, nelec, link_index=None):
        nelec = _unpack_nelec(nelec, self.spin)
        h2e = self.absorb_h1e(h1e, eri, norb, nelec, .5)
        ci1 = self.contract_2e(h2e, fcivec, norb, nelec, link_index)
        return numpy.dot(fcivec.reshape(-1), ci1.reshape(-1))

    def spin_square(self, fcivec, norb, nelec):
        nelec = _unpack_nelec(nelec, self.spin)
        return spin_op.spin_square0(fcivec, norb, nelec)
    spin_square.__doc__ = spin_op.spin_square0.__doc__

    @lib.with_doc(make_rdm1s.__doc__)
    def make_rdm1s(self, fcivec, norb, nelec, link_index=None):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm1s(fcivec, norb, nelec, link_index)

    @lib.with_doc(make_rdm1.__doc__)
    def make_rdm1(self, fcivec, norb, nelec, link_index=None):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm1(fcivec, norb, nelec, link_index)

    @lib.with_doc(make_rdm12s.__doc__)
    def make_rdm12s(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm12s(fcivec, norb, nelec, link_index, reorder)

    @lib.with_doc(make_rdm12.__doc__)
    def make_rdm12(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm12(fcivec, norb, nelec, link_index, reorder)

    @lib.with_doc(make_rdm123s.__doc__)
    def make_rdm123s(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm123s(fcivec, norb, nelec, link_index, reorder)

    @lib.with_doc(make_rdm123.__doc__)
    def make_rdm123(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm123(fcivec, norb, nelec, link_index, reorder)

    @lib.with_doc(make_rdm1234s.__doc__)
    def make_rdm1234s(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm1234s(fcivec, norb, nelec, link_index, reorder)

    @lib.with_doc(make_rdm1234.__doc__)
    def make_rdm1234(self, fcivec, norb, nelec, link_index=None, reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return make_rdm1234(fcivec, norb, nelec, link_index, reorder)

    def make_rdm2(self, fcivec, norb, nelec, link_index=None, reorder=True):
        r'''Spin traced 2-particle density matrice

        NOTE the 2pdm is :math:`\langle p^\dagger q^\dagger s r\rangle` but
        stored as [p,r,q,s]
        '''
        nelec = _unpack_nelec(nelec, self.spin)
        return self.make_rdm12(fcivec, norb, nelec, link_index, reorder)[1]

    @lib.with_doc(trans_rdm1s.__doc__)
    def trans_rdm1s(self, cibra, ciket, norb, nelec, link_index=None):
        nelec = _unpack_nelec(nelec, self.spin)
        return trans_rdm1s(cibra, ciket, norb, nelec, link_index)

    @lib.with_doc(trans_rdm1.__doc__)
    def trans_rdm1(self, cibra, ciket, norb, nelec, link_index=None):
        nelec = _unpack_nelec(nelec, self.spin)
        return trans_rdm1(cibra, ciket, norb, nelec, link_index)

    @lib.with_doc(trans_rdm12s.__doc__)
    def trans_rdm12s(self, cibra, ciket, norb, nelec, link_index=None,
                     reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return trans_rdm12s(cibra, ciket, norb, nelec, link_index, reorder)

    @lib.with_doc(trans_rdm12.__doc__)
    def trans_rdm12(self, cibra, ciket, norb, nelec, link_index=None,
                    reorder=True):
        nelec = _unpack_nelec(nelec, self.spin)
        return trans_rdm12(cibra, ciket, norb, nelec, link_index, reorder)

    def large_ci(self, fcivec, norb, nelec,
                 tol=getattr(__config__, 'fci_addons_large_ci_tol', .1),
                 return_strs=getattr(__config__, 'fci_addons_large_ci_return_strs', True)):
        nelec = _unpack_nelec(nelec, self.spin)
        return addons.large_ci(fcivec, norb, nelec, tol, return_strs)

    def contract_ss(self, fcivec, norb, nelec):  # noqa: F811
        nelec = _unpack_nelec(nelec, self.spin)
        return contract_ss(fcivec, norb, nelec)

    def gen_linkstr(self, norb, nelec, tril=True, spin=None):
        if spin is None:
            spin = self.spin
        neleca, nelecb = _unpack_nelec(nelec, spin)
        if tril:
            link_indexa = cistring.gen_linkstr_index_trilidx(range(norb), neleca)
            link_indexb = cistring.gen_linkstr_index_trilidx(range(norb), nelecb)
        else:
            link_indexa = cistring.gen_linkstr_index(range(norb), neleca)
            link_indexb = cistring.gen_linkstr_index(range(norb), nelecb)
        return link_indexa, link_indexb


class FCISolver(FCIBase):
    # transform_ci_for_orbital_rotation only available for FCI wavefunctions.
    # Some approx FCI solver does not have this functionality.
    def transform_ci_for_orbital_rotation(self, fcivec, norb, nelec, u):
        nelec = _unpack_nelec(nelec, self.spin)
        return addons.transform_ci_for_orbital_rotation(fcivec, norb, nelec, u)

    to_gpu = lib.to_gpu

FCI = FCISolver

class FCIvector(numpy.ndarray):
    '''An 2D np array for FCI coefficients'''

    # Special cases for ndarray when the array was modified (through ufunc)
    def __array_wrap__(self, out):
        if out.shape == self.shape:
            return out
        elif out.shape == ():  # if ufunc returns a scalar
            return out[()]
        else:
            return out.view(numpy.ndarray)

def _unpack(norb, nelec, link_index, spin=None):
    if link_index is None:
        neleca, nelecb = _unpack_nelec(nelec, spin)
        link_indexa = link_indexb = cistring.gen_linkstr_index_trilidx(range(norb), neleca)
        if neleca != nelecb:
            link_indexb = cistring.gen_linkstr_index_trilidx(range(norb), nelecb)
        return link_indexa, link_indexb
    else:
        return link_index



from . import mp, ci, cc, mcscf, fci, tdscf
try:
    from . import doci
except ImportError:
    pass

# Note the agf2 module implicitly import mpi4py. This module should not be
# automatically imported until the dependency to mpi4py is completely removed.



#!/usr/bin/env python
# Copyright 2014-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Matthew Hennefath <mhennefarth@uchicago.edu>

import h5py
from pyscf.lib.chkfile import load_mol, load, dump
from pyscf.lib import H5FileWrap
from pyscf.mcscf.addons import StateAverageMixFCISolver


def load_pdft(chkfile):
    return load_mol(chkfile), load(chkfile, "pdft")


def dump_mcpdft(
    mc,
    chkfile=None,
    key="pdft",
    e_tot=None,
    e_ot=None,
    e_states=None,
    e_mcscf=None,
    mcscf_key="mcscf",
):
    """Save MC-PDFT calculation results in chkfile"""
    if chkfile is None:
        chkfile = mc.chkfile
    if e_tot is None:
        e_tot = mc.e_tot
    if e_ot is None:
        e_ot = mc.e_ot

    if h5py.is_hdf5(chkfile):
        mode = "a"
    else:
        mode = "w"

    with H5FileWrap(chkfile, mode) as fh5:
        if mode == "a" and key in fh5:
            del fh5[key]

        def store(subkey, val):
            if val is not None:

                fh5[key + "/" + subkey] = val

        store("e_tot", e_tot)
        store("e_ot", e_ot)
        store("e_states", e_states)
        store("e_mcscf", e_mcscf)

        # Now lets link
        if mcscf_key in fh5:
            hard_links = {
                "mo_coeff": "mo_coeff",
                "ci__from_list__": "ci__from_list__",
                "ci": "ci",
            }

            for s, d in hard_links.items():
                if s in fh5[mcscf_key]:
                    fh5[key + "/" + d] = fh5[mcscf_key + "/" + s]


def dump_lpdft(
    mc,
    chkfile=None,
    key="pdft",
    e_tot=None,
    e_states=None,
    e_mcscf=None,
    ci=None,
    mcscf_key="mcscf",
):
    """Save L-PDFT calculation results in chkfile"""
    if chkfile is None:
        chkfile = mc.chkfile

    if e_tot is None:
        e_tot = mc.e_tot

    if e_states is None:
        e_states = mc.e_states

    if e_mcscf is None:
        e_mcscf = mc.e_mcscf

    if h5py.is_hdf5(chkfile):
        mode = "a"

    else:
        mode = "w"

    mixed_ci = isinstance(mc.fcisolver, StateAverageMixFCISolver) and ci is not None

    with H5FileWrap(chkfile, mode) as fh5:
        if mode == "a" and key in fh5:
            del fh5[key]

        def store(subkey, val):
            if val is not None:
                fh5[key + "/" + subkey] = val

        store("e_tot", e_tot)
        store("e_states", e_states)
        store("e_mcscf", e_mcscf)
        if not mixed_ci:
            store("ci", ci)

        if mcscf_key in fh5:
            hard_links = {"mo_coeff": "mo_coeff"}
            for s, d in hard_links.items():
                if s in fh5[mcscf_key]:
                    fh5[key + "/" + d] = fh5[mcscf_key + "/" + s]

    if mixed_ci:
        dump(chkfile, "pdft/ci", ci)



#!/usr/bin/env python
# Copyright 2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Nikolay A. Bogdanov <n.bogdanov@inbox.lv>
#

from functools import reduce
import numpy
import libmsym as msym
from pyscf.lib import param
from pyscf.lib import logger
from pyscf.data import elements
from pyscf import symm
from pyscf.lo.orth import vec_lowdin


def sph_mval(label):
    '''Get integer l_z value of given spherical AO label'''
    m_list = {'': 0,
              'x': 1, 'y': -1, 'z': 0,
              'xy': -2, 'yz': -1, 'z^2': 0, 'xz': 1, 'x2-y2': 2}
    try:
        m = int(label)
    except(ValueError):
        m = m_list[label]
    return m


def basis_info(ao_labels):
    info = [[id, int(nl[:-1]), param.ANGULARMAP[nl[-1]], sph_mval(m_label)]
            for (id, elem, nl, m_label) in ao_labels]
    return info


def msym_basis(ao_labels, msym_atoms):
    '''Generate list of basis functions for libmsym from pyscf data

    Given pyscf Mole object mol use as follows:
    msym_basis(mol.ao_labels(fmt=False), msym_atoms(mol._atom))
    '''
    basis_functions = [msym.RealSphericalHarmonic(element=msym_atoms[id],
                                                  n=int(nl[:-1]),
                                                  l=param.ANGULARMAP[nl[-1]],
                                                  m=sph_mval(m_label))
                       for (id, elem, nl, m_label) in ao_labels]
    return basis_functions


def msym_atoms(_atom):
    '''Generate list of atoms for libmsym from pyscf mol._atom'''
    atoms = [msym.Element(coordinates=coord,
                          charge=elements.charge(elem))
             for elem, coord in _atom]
    return atoms


def log_symmetry_info(mol):
    if mol.topgroup == mol.groupname:
        logger.info(mol, 'point group symmetry = %s', mol.topgroup)
    else:
        logger.info(mol, 'point group symmetry = %s, use subgroup %s',
                    mol.topgroup, mol.groupname)
    for ir in range(mol.symm_orb.__len__()):
        logger.info(mol, 'num. orbitals of irrep %s = %d',
                    mol.irrep_name[ir], mol.symm_orb[ir].shape[1])


def d2h_compat_irrep_id(mol, msym_symm_orb):
    '''Generate irrep id for non-Abelian SALCs compatible with the highest Abelian subgroup

    Mole.irrep_id is not just list of irrep indices, but also contains information of symmetry descent,
    such that irrep_id % 10 gives irrep_id of the highest Abelian subgroup detected for the mol.
    '''
    msym_irrep_id = []
    abelian_irrep_id_dict = {}
    for block_salc in msym_symm_orb:
        block_abelian_id = numpy.unique(symm.label_orb_symm(mol, mol.irrep_id, mol.symm_orb, block_salc))
        if len(block_abelian_id) == 0:
            raise RuntimeError('No Abelian id found for SALCs generated by libmsym')
        if len(block_abelian_id) > 1:
            raise NotImplementedError('There is no unique mapping from non-Abelian to Abelian subgroup')
        d2h_id = block_abelian_id[0]
        if d2h_id not in abelian_irrep_id_dict.keys():
            msym_irrep_id.append(d2h_id)
            abelian_irrep_id_dict[d2h_id] = 1
        else:
            msym_irrep_id.append(d2h_id + 10*abelian_irrep_id_dict[d2h_id])
            abelian_irrep_id_dict[d2h_id] += 1
    return msym_irrep_id


def gen_mol_msym(mol, tol=1e-14, verbose=True):
    '''Genetate Mole object supporting non-Abelian point groups

    Produce an instance of :class:`Mole` with symmetry-adapded AO (known as
    symmetry-adapded linear combinations, SALCs) generated by libmsym
    (https://github.com/mcodev31/libmsym) library, which supports real
    spherical harmonics with any angular momentum for point groups with real
    characters (Ci, Cs, Cnv, Dn, Dnh, Dnd, Td, O, Oh, I and Ih).

    Args:
        mol : an instance of :class:`Mole`

    Kwargs:
        tol : float
            Tolerance for symmetry detection
        verbose : bool
            Whether to write output about symmetry detection

    Returns:
        an instance of :class:`Mole` with symmetry information generated by libmsym
    '''
    mol_msym = mol.copy()
    if verbose:
        logger.info(mol_msym, '\n*** Use libmsym to generate symmetry-adapted AOs ***')
        logger.info(mol_msym, 'tolerance used to detect symmetry = %s', tol)
    if mol.symmetry is False:
        raise RuntimeError('To use libmsym interface, symmetry has to be enabled in the parent Mole object')
    mol_msym.abelian_groupname = mol.groupname
    if mol.cart:
        raise NotImplementedError('libmsym does not support cartesian harmonics')
    prop_atoms = mol_msym.format_atom(mol_msym._atom,
                                      mol_msym._symm_orig,
                                      mol_msym._symm_axes,
                                      'Angstrom')
    atoms_msym_fmt = msym_atoms(prop_atoms)
    with msym.Context(elements=atoms_msym_fmt,
                      basis_functions=msym_basis(mol_msym.ao_labels(fmt=False),
                                                 atoms_msym_fmt)) as ctx:
        ctx.set_thresholds(zero=tol, geometry=tol, angle=tol,
                           equivalence=tol, eigfact=tol, permutation=tol,
                           orthogonalization=tol)
        mol_msym.groupname = ctx.find_symmetry()
        # character_table = ctx.character_table.table
        irrep_name = [irrep.name for irrep in ctx.character_table.symmetry_species]
        irrep_dim = [irrep.dim for irrep in ctx.character_table.symmetry_species]
        (salc_mo_coeff, salc_symm_id, salc_partner_func) = ctx.salcs
        salc_mo_coeff = salc_mo_coeff.T
        salc_partner_id = [pf.dim for pf in salc_partner_func]
        # partners_by_irrep = [(id, numpy.sort(numpy.unique(salc_partner_id[salc_symm_id==id])))
        #                      for id in numpy.unique(salc_symm_id)]
        labels = numpy.array([irrep_name[idx]
                              + ("_" + str(pf))*(irrep_dim[idx] > 1)  # add partner id for dim > 1
                              for idx, pf in zip(salc_symm_id, salc_partner_id)])
        mol_msym.irrep_name = []
        for lab in labels:
            if lab not in mol_msym.irrep_name:
                mol_msym.irrep_name.append(lab)
        mol_msym.symm_orb = [salc_mo_coeff[:, labels == i_label] for i_label in mol_msym.irrep_name]
        # in-place rotate SALCs to the original geometry
        l_idx = symm.basis.ao_l_dict(mol_msym)
        Ds = symm.basis._momentum_rotation_matrices(mol_msym, mol_msym._symm_axes)
        for c_ir in mol_msym.symm_orb:
            nso = c_ir.shape[1]
            for l, idx in l_idx.items():
                c = c_ir[idx].reshape(-1, Ds[l].shape[1], nso)
                c_ir[idx] = numpy.einsum('nm,smp->snp', Ds[l], c).reshape(-1, nso)
        mol_msym.irrep_id = d2h_compat_irrep_id(mol, mol_msym.symm_orb)
    if verbose:
        log_symmetry_info(mol_msym)
    return mol_msym


def symmetrize_space_pyscf(mol, mo, s=None, orthonormalize=False):
    '''Symmetrize MOs with pyscf function but symmetry generated by libmsym'''
    mol_msym = gen_mol_msym(mol)
    if s is None:
        s = mol.intor_symmetric('int1e_ovlp')
    mo_msym = symm.symmetrize_space(mol_msym, mo, s=s)
    if orthonormalize:
        mo_msym = vec_lowdin(mo_msym, s=s)
    return mo_msym


def symmetrize_space_libmsym(mol, mo, tol=1e-14, check=True, orthonormalize=False):
    '''Symmetrize MOs with libmsym itself'''
    if check:
        s = mol.intor_symmetric('int1e_ovlp')
        nmo = mo.shape[1]
        # s_mo = numpy.dot(s, mo_msym)
        max_non_orth = abs(reduce(numpy.dot, (mo.conj().T, s, mo)) - numpy.eye(nmo)).max()
        if max_non_orth > tol:
            if orthonormalize:
                print('Input orbitals are not orthogonalized: '+str(max_non_orth))
                mo = vec_lowdin(mo, s=s)
                max_non_orth = abs(reduce(numpy.dot, (mo.conj().T, s, mo)) - numpy.eye(nmo)).max()
                print("Input orbitals were orthonormalized: " + str(max_non_orth))
            else:
                raise ValueError('Input orbitals are not orthogonalized: '+str(max_non_orth))
    # mol_msym = gen_mol_msym(mol)
    my_msym_atoms = msym_atoms(mol._atom)
    with msym.Context(elements=my_msym_atoms,
                      basis_functions=msym_basis(mol.ao_labels(fmt=False), my_msym_atoms)) as ctx:
        ctx.set_thresholds(zero=tol, geometry=tol, angle=tol,
                           equivalence=tol, eigfact=tol, permutation=tol,
                           orthogonalization=tol)
        ctx.find_symmetry()
        sym_mo_coeff, sym_symm_id, symm_partner_func = ctx.symmetrize_wavefunctions(mo.T)
    mo_msym = sym_mo_coeff.T
    if check:
        nmo = mo_msym.shape[1]
        max_non_orth = abs(
            reduce(numpy.dot, (mo_msym.conj().T, s, mo_msym)) - numpy.eye(nmo)).max()
        if max_non_orth > tol:
            if orthonormalize:
                mo_msym = vec_lowdin(mo_msym, s=s)
                max_non_orth = abs(
                    reduce(numpy.dot, (mo_msym.conj().T, s, mo_msym)) - numpy.eye(nmo)).max()
            else:
                raise ValueError('Output orbitals are not orthogonalized: ' + str(max_non_orth))
    return mo_msym


symmetrize_space = symmetrize_space_pyscf



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import ctypes
import numpy as np
from pyscf import __config__
from pyscf import lib, gto
from pyscf.lib import logger
from pyscf.pbc import tools
from pyscf.pbc.gto import pseudo
from pyscf.pbc.gto.pseudo import pp_int
from pyscf.pbc.lib.kpts_helper import gamma_point
from pyscf.pbc.lib.kpts import KPoints
from pyscf.pbc.df import fft
from pyscf.pbc.dft.multigrid import _backend_c as backend

PP_WITH_RHO_CORE = getattr(__config__, 'pbc_dft_multigrid_pp_with_rho_core', True)


def make_rho_core(cell, mesh=None, precision=None, atm_id=None):
    if mesh is None:
        mesh = cell.mesh
    fakecell, max_radius = fake_cell_vloc_part1(cell, atm_id=atm_id, precision=precision)

    a = cell.lattice_vectors()
    b = np.linalg.inv(a.T)

    rho_core = backend.build_core_density(
        fakecell._atm,
        fakecell._bas,
        fakecell._env,
        mesh,
        cell.dimension,
        a,
        b,
        max_radius,
    )
    logger.debug(cell, 'Number of core electrons: %.9f',
                 -np.sum(rho_core) * cell.vol / np.prod(mesh))
    return rho_core


def _get_pp_without_erf(mydf, kpts=None):
    '''Get the periodic pseudopotential nuc-el AO matrix, with G=0 removed.
    '''
    cell = mydf.cell
    if kpts is None:
        kpts = np.zeros((1,3))
    elif isinstance(kpts, KPoints):
        kpts = kpts.kpts_ibz
    kpts, is_single_kpt = fft._check_kpts(mydf, kpts)

    vpp = pp_int.get_pp_loc_part2(cell, kpts)
    vppnl = pp_int.get_pp_nl(cell, kpts)

    for k, kpt in enumerate(kpts):
        if gamma_point(kpt):
            vpp[k] = vpp[k].real + vppnl[k].real
        else:
            vpp[k] += vppnl[k]
    vppnl = None

    if is_single_kpt:
        vpp = vpp[0]
    return np.asarray(vpp)


def get_pp_loc_part1_gs(cell, Gv):
    coulG = tools.get_coulG(cell, Gv=Gv)
    G2 = np.einsum('ix,ix->i', Gv, Gv)
    G0idx = np.where(G2==0)[0][0]

    coords = cell.atom_coords()

    natm = cell.natm
    Z = np.empty(natm)
    rloc = np.empty(natm)
    for ia in range(natm):
        Z[ia] = cell.atom_charge(ia)
        symb = cell.atom_symbol(ia)
        if symb in cell._pseudo:
            rloc[ia] = cell._pseudo[symb][1]
        else:
            rloc[ia] = -999

    out = backend.pp_loc_part1_gs(
            coulG, Gv, G2, G0idx,
            Z, coords, rloc)
    return out


def _get_vpplocG_part1(mydf, with_rho_core=PP_WITH_RHO_CORE):
    cell = mydf.cell
    mesh = mydf.mesh

    if not with_rho_core:
        # compute rho_core directly in G-space
        # this is much slower that the following
        Gv = cell.get_Gv(mesh)
        vpplocG_part1 = get_pp_loc_part1_gs(cell, Gv)
    else:
        # compute rho_core in real space then transform to G-space
        weight = cell.vol / np.prod(mesh)
        rho_core = make_rho_core(cell)
        rhoG_core = weight * tools.fft(rho_core, mesh)
        rho_core = None
        coulG = tools.get_coulG(cell, mesh=mesh)
        vpplocG_part1 = rhoG_core * coulG
        rhoG_core = coulG = None
        # G = 0 contribution
        chargs = cell.atom_charges()
        symbs = map(cell.atom_symbol, range(cell.natm))
        rloc = [cell._pseudo[symb][1] if symb in cell._pseudo else 0.
                for symb in symbs]
        rloc = np.asarray(rloc)
        vpplocG_part1[0] += 2. * np.pi * np.sum(rloc * rloc * chargs)
    return vpplocG_part1


def get_vpploc_part1_ip1(mydf, kpts=np.zeros((1,3))):
    from .multigrid_pair import _get_j_pass2_ip1
    if isinstance(kpts, KPoints):
        raise NotImplementedError
    vG = mydf.vpplocG_part1
    if vG is None:
        vG = _get_vpplocG_part1(mydf)

    vpp_kpts = _get_j_pass2_ip1(mydf, vG, kpts, hermi=0, deriv=1)
    if gamma_point(kpts):
        vpp_kpts = vpp_kpts.real
    if len(kpts) == 1:
        vpp_kpts = vpp_kpts[0]
    return vpp_kpts


def vpploc_part1_nuc_grad(mydf, dm, kpts=np.zeros((1,3)), atm_id=None, precision=None):
    from .multigrid_pair import _eval_rhoG
    if isinstance(kpts, KPoints):
        raise NotImplementedError
    t0 = (logger.process_clock(), logger.perf_counter())
    cell = mydf.cell
    fakecell, max_radius = fake_cell_vloc_part1(cell, atm_id=atm_id, precision=precision)
    atm = fakecell._atm
    bas = fakecell._bas
    env = fakecell._env

    a = cell.lattice_vectors()
    b = np.linalg.inv(a.T)

    mesh = mydf.mesh
    ngrids = np.prod(mesh)
    comp = 3

    if mydf.rhoG is None:
        rhoG = _eval_rhoG(mydf, dm, hermi=1, kpts=kpts, deriv=0)
    else:
        rhoG = mydf.rhoG
    rhoG = rhoG[...,0,:]
    rhoG = rhoG.reshape(-1,ngrids)
    if rhoG.shape[0] == 2: #unrestricted
        rhoG = rhoG[0] + rhoG[1]
    else:
        assert rhoG.shape[0] == 1
        rhoG = rhoG[0]

    coulG = tools.get_coulG(cell, mesh=mesh)
    vG = np.multiply(rhoG, coulG)
    v_rs = tools.ifft(vG, mesh).real

    grad = backend.int_gauss_charge_v_rs(
        v_rs,
        comp,
        atm,
        bas,
        env,
        mesh,
        cell.dimension,
        a,
        b,
        max_radius,
    )
    grad *= -1
    t0 = logger.timer(mydf, 'vpploc_part1_nuc_grad', *t0)
    return grad


def fake_cell_vloc_part1(cell, atm_id=None, precision=None):
    '''
    Generate fakecell for the long-range term of the local part of
    the GTH pseudo-potential. Also stores the atomic radii.
    Differs from pp_int.fake_cell_vloc(cell, cn=0) in the normalization factors.
    '''
    from pyscf.pbc.gto.cell import pgf_rcut
    if atm_id is None:
        atm_id = np.arange(cell.natm)
    else:
        atm_id = np.asarray(atm_id)
    natm = len(atm_id)

    if precision is None:
        precision = cell.precision

    max_radius = 0
    kind = {}
    # FIXME prec may be too tight
    prec = precision ** 2
    for symb in cell._pseudo:
        charge = np.sum(cell._pseudo[symb][0])
        rloc = cell._pseudo[symb][1]
        zeta = .5 / rloc**2
        norm = (zeta / np.pi) ** 1.5
        radius = pgf_rcut(0, zeta, charge*norm, precision=prec)
        max_radius = max(radius, max_radius)
        kind[symb] = [zeta, norm, radius]

    fake_env = [cell.atom_coords()[atm_id].ravel()]
    fake_atm = cell._atm[atm_id].copy().reshape(natm,-1)
    fake_atm[:,gto.PTR_COORD] = np.arange(0, natm*3, 3)
    ptr = natm * 3
    fake_bas = []
    for ia, atm in enumerate(atm_id):
        if cell.atom_charge(atm) == 0:  # pass ghost atoms
            continue

        symb = cell.atom_symbol(atm)
        if symb in kind:
            fake_env.append(kind[symb])
        else:
            alpha = 1e16
            norm = (alpha / np.pi) ** 1.5
            radius = 0.0
            fake_env.append([alpha, norm, radius])
        fake_bas.append([ia, 0, 1, 1, 0, ptr, ptr+1, 0])
        fake_atm[ia,gto.PTR_RADIUS] = ptr+2
        ptr += 3

    fakecell = cell.copy(deep=False)
    fakecell._atm = np.asarray(fake_atm, order="C", dtype=np.int32)
    fakecell._bas = np.asarray(fake_bas, order="C", dtype=np.int32).reshape(-1, gto.BAS_SLOTS)
    fakecell._env = np.asarray(np.hstack(fake_env), order="C", dtype=float)
    return fakecell, max_radius
